# For more information see:
# https://docs.freebsd.org/en/articles/vinum/
# https://www.usenix.org/legacy/events/usenix99/full_papers/lehey/lehey.pdf
from __future__ import annotations

import io
import logging
import os
from collections import defaultdict
from functools import cached_property
from typing import TYPE_CHECKING, BinaryIO, TypedDict, TypeVar

from dissect.util import ts

from dissect.volume.raid.raid import (
    RAID,
    Configuration,
    PhysicalDisk,
    VirtualDisk,
)
from dissect.volume.raid.stream import Layout, Level
from dissect.volume.vinum.c_vinum import MAGIC_ACTIVE, MAGIC_INACTIVE, c_vinum
from dissect.volume.vinum.config import (
    PlexOrg,
    PlexState,
    SDState,
    parse_vinum_config,
)

if TYPE_CHECKING:
    from dissect.volume.raid.raid import (
        DiskMap,
    )
    from dissect.volume.vinum.config import (
        SD,
        Plex,
        Volume,
    )

    VinumPhysicalDiskDescriptor = BinaryIO | "VinumPhysicalDisk"

log = logging.getLogger(__name__)
log.setLevel(os.getenv("DISSECT_LOG_VINUM", "CRITICAL"))


class Vinum(RAID):
    """Read Vinum RAID sets of one or multiple devices/file-like objects.

    Use this class to read from Vinum RAID sets.

    A single Vinum RAID set is defined by a Volume in the Vinum configuration.
    This configuration is present on all physical disks and contains
    information on all the RAID sets in the system A Vinum Volume can have one
    or more Plexes.

    A Plex can be thought of as one of the individual disks in a mirrored
    array. A Plex can contain one or more Vinum SDs. The Plex defines the type
    of RAID in which these SDs are organized.

    An SD contains information about the actual physical disk and points to the
    device of this disk.

    Args:
        fh: A single file-like object or :class:`VinumPhysicalDisk`, or a list
            of multiple belonging to the same RAID set.
    """

    def __init__(self, fh: list[VinumPhysicalDiskDescriptor] | VinumPhysicalDiskDescriptor):
        fhs = [fh] if not isinstance(fh, list) else fh
        physical_disks = [VinumPhysicalDisk(fh) if not isinstance(fh, VinumPhysicalDisk) else fh for fh in fhs]

        super().__init__([VinumConfiguration(physical_disks)])


T = TypeVar("T")
ByName = dict[bytes, T]
DefaultByName = defaultdict[bytes, T]


class Config(TypedDict):
    """Configuration."""

    volumes: ByName[Volume]
    plexes: ByName[Plex]
    sds: ByName[SD]


class VinumConfiguration(Configuration):
    def __init__(self, physical_disks: list[VinumPhysicalDisk]):
        # These hold the most recent config for each volume/plex/sd
        config: Config = {"volumes": {}, "plexes": {}, "sds": {}}
        disks_by_name: ByName[VinumPhysicalDisk] = {}

        # Find the most recent configuration for each Volume/Plex/SD by merging
        # all configs from all physical disks.
        for disk in physical_disks:
            if not disk.active:
                # Assuming here that if a disk is marked as inactive, it's
                # configuration is also old and/or possibly inaccurate.
                continue

            disks_by_name[disk.id] = disk

            # For now we only use the first config block. The second config block
            # could be useful/used as a fallback if for instance the first one
            # fails to parse.
            #
            # The self.header.label.last_update timestamp is used to see if this
            # disk contains the latest/newest config of the disks in a set. The
            # disk is ignored if there is no header (which in our case would have
            # raised a ValueError) or if the state is not set to GV_DRIVE_UP (an
            # internal kernel state which we don't have and thus can ignore).
            #
            # Plexes with a non-existing Volume and SDs with a non-existing Plex or
            # Drive (VinumPhysicalDisk) can not be used. They result in an error in
            # the FreeBSD kernel code.
            config_data = parse_vinum_config(disk.config_time, disk.config)

            for config_type, new_items in config_data.items():
                cur_config = config[config_type]

                for new_item in new_items:
                    if cur_item := cur_config.get(new_item.name):
                        if new_item.timestamp > cur_item.timestamp:
                            cur_config[new_item.name] = new_item
                    else:
                        cur_config[new_item.name] = new_item

        # plexes_by_name contains all *active* plexes
        plexes_by_name: ByName[Plex] = {}

        # plexes_by_volume contains all *active* plexes grouped by the name of
        # the volume they belong to
        plexes_by_volume: DefaultByName[list[Plex]] = defaultdict(list)
        # sds_by_plex_by_volume contains all sds for all *active* plexes
        # grouped by plex and volume name
        sds_by_plex_by_volume: DefaultByName[DefaultByName[list[SD]]] = defaultdict(lambda: defaultdict(list))

        # Check whether the found plexes are viable and filter them out if they
        # are not.
        # Note that if a complete Volume is down, we still try to use it (we don't check on Volume.state).
        for plex in config["plexes"].values():
            if config["volumes"].get(plex.volume):
                # A plex in the DOWN state is probably not fit to use
                if plex.state != PlexState.DOWN:
                    if plex.org != PlexOrg.DISORG:
                        plexes_by_name[plex.name] = plex
                        plexes_by_volume[plex.volume].append(plex)
                    else:
                        log.warning("Plex %r has an unknown organisation, ignoring plex", plex.name)
                else:
                    log.warning("Plex %r is down, ignoring plex", plex.name)
            else:
                log.warning("Unknown volume %r for plex %r, ignoring plex", plex.volume, plex.name)

        # Check whether the found sds are viable and filter them out if they
        # are not.
        for sd in config["sds"].values():
            if plex := plexes_by_name.get(sd.plex):
                sds_by_plex_by_volume[plex.volume][sd.plex].append(sd)
            else:
                log.warning("Unknown or inactive plex %r for sd %r, ignoring sd", sd.plex, sd.name)

        # The construction of the disk_map assumes the configuration
        # information is complete and no sd configs are missing.
        # Note that there are no checks done on the completeness of the sd
        # values, these can theoretically be None due to a corrupt
        # configuration.
        disk_map_by_plex_by_volume: DefaultByName[DefaultByName[list[SD]]] = defaultdict(lambda: defaultdict(list))
        for volume_id, sds_by_plex in sds_by_plex_by_volume.items():
            for plex_id, sds in sds_by_plex.items():
                sds = sorted(sds, key=lambda sd: sd.plexoffset)
                disk_map = {}
                for idx, sd in enumerate(sds):
                    if (
                        sd.state != SDState.DOWN and sd.state != SDState.DEGRADED and sd.state != SDState.INITIALIZING
                    ):  # sd's in these states are probably not fit to use
                        if sd.drive in disks_by_name:
                            disk_map[idx] = (0, disks_by_name[sd.drive])
                        else:
                            log.warning("Physical disk %r for sd %r is missing, ignoring sd", sd.drive, sd.name)
                    else:
                        log.warning("SD %r is not in a usable state: %r, ignoring sd", sd.name, sd.state)
                if disk_map:
                    disk_map_by_plex_by_volume[volume_id][plex_id] = disk_map

        # Each volume represents a separate virtual disk
        virtual_disks = []
        for volume_id, plexes in plexes_by_volume.items():
            volume = config["volumes"][volume_id]
            if plexes:
                # Special case if there is only 1 plex (no mirroring)
                if len(plexes) == 1:
                    plex = plexes[0]
                    disk_map = disk_map_by_plex_by_volume[volume.name][plex.name]
                    if disk_map:
                        sds = sds_by_plex_by_volume[volume.name][plex.name]
                        virtual_disks.append(VinumPlexDisk(volume, plex, sds, disk_map))
                else:
                    plex_map = {}
                    # There is no official order in the plexes. However if they
                    # are named automatically, they have a pseudo order due to
                    # there names being constructed with p0, p1, etc.
                    splexes = sorted(plexes, key=lambda plex: plex.name)
                    for idx, plex in enumerate(splexes):
                        disk_map = disk_map_by_plex_by_volume[volume.name][plex.name]
                        if disk_map:
                            sds = sds_by_plex_by_volume[volume.name][plex.name]
                            plex_disk = VinumPlexDisk(volume, plex, sds, disk_map)
                            plex_map[idx] = (0, plex_disk)

                    if plex_map:
                        virtual_disks.append(VinumMirrorDisk(volume, plexes, plex_map))

            else:
                log.warning("Volume %r has no or only inactive plexes, ignoring volume", volume.name)

        if not virtual_disks:
            raise ValueError(
                "Invalid vinum raid configuration, no volumes found with an active and complete set of disks"
            )

        super().__init__(physical_disks, virtual_disks)


org_to_level = {
    PlexOrg.CONCAT: Level.LINEAR,
    PlexOrg.STRIPED: Level.RAID0,
    PlexOrg.RAID5: Level.RAID5,
}

org_to_layout = {
    PlexOrg.CONCAT: 0,
    PlexOrg.STRIPED: 0,
    PlexOrg.RAID5: Layout.LEFT_ASYMMETRIC,
}


class VinumPlexDisk(VirtualDisk):
    def __init__(
        self,
        volume: Volume,
        plex: Plex,
        sds: list[SD],
        disk_map: DiskMap,
    ):
        self.volume = volume
        self.plex = plex
        self.sds = sds

        if (level := org_to_level.get(plex.org)) is None:
            raise ValueError(f"Plex {plex.name} has an unsupported RAID level: {plex.org}")
        if (layout := org_to_layout.get(plex.org)) is None:
            raise ValueError(f"Plex {plex.name} has an unsupported RAID level: {plex.org}")

        stripe_size = 0 if plex.org == PlexOrg.CONCAT else plex.stripesize  # concatenated disks don't have stripes

        size = 0
        sd = sds[0]
        if plex.org == PlexOrg.RAID5:
            # SDs in a vinum RAID5 org are required to have equal size, so we
            # don't need to determine the smallest disk.
            size = (len(sds) - 1) * sd.length
        elif plex.org == PlexOrg.STRIPED:
            # SDs in a vinum STRIPED org are required to have equal size, so we
            # can just multiply
            size = len(sds) * sd.length
        else:
            for sd in sds:
                size += sd.length

        super().__init__(
            volume.name.decode(errors="backslashreplace"),
            plex.name.decode(errors="surrogateescape"),
            size,
            level,
            layout,
            stripe_size,
            len(sds),
            disk_map,
        )


class VinumMirrorDisk(VirtualDisk):
    def __init__(
        self,
        volume: Volume,
        plexes: list[Plex],
        plex_map: DiskMap,
    ):
        self.volume = volume
        self.plexes = plexes

        # The VinumConfiguration class will make sure there is at least 1 plex disk.
        _, plex_disk = next(iter(plex_map.values()))
        size = plex_disk.size
        super().__init__(
            volume.name.decode(errors="backslashreplace"),
            volume.name.decode(errors="surrogateescape"),
            size,
            Level.RAID1,
            0,  # simple mirrors don't have a layout
            0,  # simple mirrors don't have a stripe size
            len(plexes),
            plex_map,
        )


class VinumPhysicalDisk(PhysicalDisk):
    """Parse config from an Vinum device.

    Args:
        fh: The file-like object to read config from.
    """

    def __init__(self, fh: BinaryIO):
        self.fh = fh

        fh.seek(c_vinum.GV_HDR_OFFSET)
        self.header = c_vinum.header(fh)

        if self.header.magic in MAGIC_ACTIVE:
            self.active = True
        elif self.header.magic in MAGIC_INACTIVE:
            self.active = False
        else:
            raise ValueError("File-like object is not a Vinum device")

        self.id = self.header.label.name.rstrip(b"\x00")
        self.name = self.header.label.name.rstrip(b"\x00").decode(errors="backslashreplace")

        last_update = self.header.label.last_update
        config_epoch = last_update.sec + last_update.usec * 1e-6
        self.config_time = ts.from_unix(config_epoch)

        size = self.header.label.drive_size
        if not size:
            fh.seek(0, io.SEEK_END)
            size = fh.tell()

        super().__init__(fh, c_vinum.GV_DATA_START, size)

    def _read_config(self, config_offset: int) -> bytes:
        self.fh.seek(config_offset)
        return self.fh.read(self.header.config_length)

    @cached_property
    def config(self) -> bytes:
        return self._read_config(c_vinum.GV_CFG_OFFSET)

    @cached_property
    def config2(self) -> bytes:
        return self._read_config(c_vinum.GV_CFG_OFFSET + c_vinum.GV_CFG_LEN)



from __future__ import annotations

from typing import TYPE_CHECKING, BinaryIO

from dissect.cstruct import cstruct

from dissect.volume.disk.partition import Partition
from dissect.volume.exceptions import DiskError

if TYPE_CHECKING:
    from collections.abc import Iterator

apm_def = """
struct partition_entry {
    char    signature[2];
    uint16  reserved_1;
    uint32  partition_count;
    uint32  partition_start;
    uint32  partition_size;
    char    partition_name[32];
    char    partition_type[32];
    uint32  data_start;
    uint32  data_size;
    uint32  partition_status;
    uint32  boot_code_start;
    uint32  boot_code_size;
    uint32  boot_loader_address;
    uint32  reserved_2;
    uint32  boot_code_entry;
    uint32  reserved_3;
    uint32  boot_code_checksum;
    char    processor_type[16];
    char    reserved_4[376];
};
"""

c_apm = cstruct(endian=">").load(apm_def)


class APM:
    """Apple Partition Map."""

    def __init__(self, fh: BinaryIO, sector_size: int = 512):
        self.fh = fh
        self.sector_size = sector_size
        self.apm = c_apm.partition_entry(fh)

        if self.apm.signature == b"ER":
            # We parsed the boot jump, the offset is conveniently in reserved_1 field
            fh.seek(self.apm.reserved_1)
            self.apm = c_apm.partition_entry(fh)

        if self.apm.signature != b"PM":
            raise DiskError(f"Invalid APM signature, expected 'PM', got {self.apm.signature!r}.")

        self._partitions_offset = fh.tell()
        self.partitions: list[Partition] = list(self._partitions())

    def _partitions(self) -> Iterator[Partition]:
        self.fh.seek(self._partitions_offset)
        for i in range(self.apm.partition_count):
            p = self.apm if i == 0 else c_apm.partition_entry(self.fh)

            yield Partition(
                disk=self,
                number=i + 1,  # partitions are 1-indexed
                offset=p.partition_start * self.sector_size,
                size=p.partition_size * self.sector_size,
                type=p.partition_type.rstrip(b"\x00").decode(),
                name=p.partition_name.rstrip(b"\x00").decode(),
                raw=p,
            )



from __future__ import annotations

import io
import operator
import struct
from typing import TYPE_CHECKING, BinaryIO
from uuid import UUID

from dissect.util import ts

from dissect.volume.md.c_md import SECTOR_SIZE, c_md
from dissect.volume.raid.raid import RAID, Configuration, PhysicalDisk, VirtualDisk
from dissect.volume.raid.stream import Level

if TYPE_CHECKING:
    import datetime


class MD(RAID):
    """Read an MD RAID set of one or multiple devices/file-like objects.

    Use this class to read from a RAID set.

    Args:
        fh: A single file-like object or :class:`MDPhysicalDisk`, or a list of multiple belonging to the same RAID set.
    """

    def __init__(self, fh: list[BinaryIO | MDPhysicalDisk] | BinaryIO | MDPhysicalDisk):
        fhs = [fh] if not isinstance(fh, list) else fh
        physical_disks = [MDPhysicalDisk(fh) if not isinstance(fh, MDPhysicalDisk) else fh for fh in fhs]

        config_map = {}
        for disk in physical_disks:
            config_map.setdefault(disk.set_uuid, []).append(disk)

        super().__init__([MDConfiguration(disks) for disks in config_map.values()])


class MDConfiguration(Configuration):
    def __init__(self, physical_disks: list[MDPhysicalDisk]):
        physical_disks = sorted(physical_disks, key=operator.attrgetter("raid_disk"))

        if len({disk.set_uuid for disk in physical_disks}) != 1:
            raise ValueError("Multiple MD sets detected, supply only the disks of a single set")

        virtual_disks = [MDVirtualDisk(physical_disks)]
        super().__init__(physical_disks, virtual_disks)


class MDVirtualDisk(VirtualDisk):
    def __init__(self, physical_disks: list[MDPhysicalDisk]):
        reference_disk = sorted(physical_disks, key=operator.attrgetter("events"), reverse=True)[0]
        disk_map = {disk.raid_disk: (0, disk) for disk in physical_disks if disk.raid_disk is not None}

        if reference_disk.level == Level.LINEAR:
            size = sum(disk.size for _, disk in disk_map.values())
        elif reference_disk.level == Level.RAID0:
            size = 0
            for _, disk in disk_map.values():
                size += disk.size & ~(reference_disk.chunk_size - 1)
        elif reference_disk.level in (Level.RAID1, Level.RAID4, Level.RAID5, Level.RAID6, Level.RAID10):
            size = reference_disk.sb.size * SECTOR_SIZE
        else:
            raise ValueError(
                "Invalid MD RAID configuration: No valid RAID level found for the reference disk, found: %d",
                reference_disk.level,
            )

        super().__init__(
            reference_disk.set_name,
            str(reference_disk.set_uuid),
            size,
            reference_disk.level,
            reference_disk.layout,
            reference_disk.chunk_size,
            reference_disk.raid_disks,
            disk_map,
        )


class MDPhysicalDisk(PhysicalDisk):
    """Parse metadata from an MD device.

    Supports 0.90 and 1.x metadata.

    Args:
        fh: The file-like object to read metadata from.
    """

    def __init__(self, fh: BinaryIO):
        sb_offset, sb_major, sb_minor = find_super_block(fh)
        if sb_offset is None:
            raise ValueError("File-like object is not an MD device")

        fh.seek(sb_offset * SECTOR_SIZE)
        if sb_major == 1:
            self.sb = c_md.mdp_superblock_1(fh)
        elif sb_major == 0:
            self.sb = c_md.mdp_super_t(fh)
        else:
            raise ValueError(f"Invalid MD version at {sb_offset:#x}: {sb_major}.{sb_minor}")

        if self.sb.major_version == 1:
            self.set_uuid = UUID(bytes_le=self.sb.set_uuid)
            self.set_name = self.sb.set_name.split(b"\x00", 1)[0].decode(errors="surrogateescape")
            self.events = self.sb.events
            self.chunk_sectors = self.sb.chunksize
            self.chunk_size = self.chunk_sectors * SECTOR_SIZE
            self.data_offset = self.sb.data_offset
            self.data_size = self.sb.data_size
            self.dev_number = self.sb.dev_number
            self.device_uuid = UUID(bytes_le=self.sb.device_uuid)

            role = self.sb.dev_roles[self.sb.dev_number]
            if role == c_md.MD_DISK_ROLE_JOURNAL:
                self.raid_disk = 0
            elif role <= c_md.MD_DISK_ROLE_MAX:
                self.raid_disk = role
            else:
                self.raid_disk = None

        else:
            self.set_uuid = UUID(bytes_le=self.sb.set_uuid0 + self.sb.set_uuid1 + self.sb.set_uuid2 + self.sb.set_uuid3)
            self.set_name = None
            self.events = (self.sb.events_hi << 32) | self.sb.events_lo
            self.chunk_size = self.sb.chunk_size
            self.chunk_sectors = self.chunk_size // SECTOR_SIZE
            self.data_offset = 0
            self.data_size = sb_offset
            self.dev_number = self.sb.this_disk.number
            self.device_uuid = None
            self.raid_disk = self.sb.disks[self.dev_number].raid_disk

        self.creation_time = _parse_ts(self.sb.ctime)
        self.update_time = _parse_ts(self.sb.ctime)
        self.level = self.sb.level
        self.layout = self.sb.layout
        self.raid_disks = self.sb.raid_disks
        self.sectors = self.data_size

        super().__init__(fh, self.data_offset * SECTOR_SIZE, self.data_size * SECTOR_SIZE)


def find_super_block(fh: BinaryIO) -> tuple[int | None, int | None, int | None]:
    """Find the super block of an MD device.

    Args:
        fh: The file-like object to search for the super block.

    Returns:
        A tuple containing the sector, major version, and minor version of the super block, or a
        tuple of ``None`` if no super block was found.
    """
    size = fh.size if (hasattr(fh, "size") and fh.size is not None) else fh.seek(0, io.SEEK_END)
    size //= SECTOR_SIZE

    # Super block can start at a couple of places, depending on version
    # Just try them all until we find one
    possible_sectors = [
        # 0.90.0
        (size & ~(c_md.MD_RESERVED_SECTORS - 1)) - c_md.MD_RESERVED_SECTORS,
        # Major version 1
        # 0: At least 8K, but less than 12K, from end of device
        size - 8 * 2,
        # 1: At start of device
        0,
        # 2: 4K from start of device.
        8,
    ]

    for sector in possible_sectors:
        if sector < 0:
            continue

        fh.seek(sector * SECTOR_SIZE)
        peek = fh.read(12)
        if len(peek) != 12:
            continue

        magic, major, minor = struct.unpack("<3I", peek)
        if magic == c_md.MD_SB_MAGIC:
            return sector, major, minor

    return None, None, None


def _parse_ts(timestamp: int) -> datetime.datetime:
    """Utility method for parsing MD timestamps.

    Lower 40 bits are seconds, upper 24 are microseconds.
    """
    seconds = timestamp & 0xFFFFFFFFFF
    micro = timestamp >> 40
    return ts.from_unix_us((seconds * 1000000) + micro)



# Generated by cstruct-stubgen
from typing import BinaryIO, Literal, TypeAlias, overload

import dissect.cstruct as __cs__

class _c_md(__cs__.cstruct):
    MD_RESERVED_BYTES: Literal[65536] = ...
    MD_RESERVED_SECTORS: Literal[128] = ...
    MD_SB_GENERIC_CONSTANT_WORDS: Literal[32] = ...
    MD_SB_GENERIC_STATE_WORDS: Literal[32] = ...
    MD_SB_GENERIC_WORDS: Literal[64] = ...
    MD_SB_PERSONALITY_WORDS: Literal[64] = ...
    MD_SB_DESCRIPTOR_WORDS: Literal[32] = ...
    MD_SB_DISKS: Literal[27] = ...
    MD_SB_DISKS_WORDS: Literal[864] = ...
    MD_SB_RESERVED_WORDS: Literal[0] = ...
    MD_DISK_FAULTY: Literal[0] = ...
    MD_DISK_ACTIVE: Literal[1] = ...
    MD_DISK_SYNC: Literal[2] = ...
    MD_DISK_REMOVED: Literal[3] = ...
    MD_DISK_CLUSTER_ADD: Literal[4] = ...
    MD_DISK_CANDIDATE: Literal[5] = ...
    MD_DISK_FAILFAST: Literal[10] = ...
    MD_DISK_WRITEMOSTLY: Literal[9] = ...
    MD_DISK_JOURNAL: Literal[18] = ...
    MD_DISK_ROLE_SPARE: Literal[65535] = ...
    MD_DISK_ROLE_FAULTY: Literal[65534] = ...
    MD_DISK_ROLE_JOURNAL: Literal[65533] = ...
    MD_DISK_ROLE_MAX: Literal[65280] = ...
    MD_SB_MAGIC: Literal[2838187772] = ...
    WriteMostly1: Literal[1] = ...
    FailFast1: Literal[2] = ...
    LEVEL_MULTIPATH: Literal[-4] = ...
    LEVEL_LINEAR: Literal[-1] = ...
    LEVEL_FAULTY: Literal[-5] = ...
    ALGORITHM_LEFT_ASYMMETRIC: Literal[0] = ...
    ALGORITHM_RIGHT_ASYMMETRIC: Literal[1] = ...
    ALGORITHM_LEFT_SYMMETRIC: Literal[2] = ...
    ALGORITHM_RIGHT_SYMMETRIC: Literal[3] = ...
    ALGORITHM_PARITY_0: Literal[4] = ...
    ALGORITHM_PARITY_N: Literal[5] = ...
    ALGORITHM_ROTATING_ZERO_RESTART: Literal[8] = ...
    ALGORITHM_ROTATING_N_RESTART: Literal[9] = ...
    ALGORITHM_ROTATING_N_CONTINUE: Literal[10] = ...
    ALGORITHM_LEFT_ASYMMETRIC_6: Literal[16] = ...
    ALGORITHM_RIGHT_ASYMMETRIC_6: Literal[17] = ...
    ALGORITHM_LEFT_SYMMETRIC_6: Literal[18] = ...
    ALGORITHM_RIGHT_SYMMETRIC_6: Literal[19] = ...
    ALGORITHM_PARITY_0_6: Literal[20] = ...
    ALGORITHM_PARITY_N_6: Literal[5] = ...
    class mdp_device_descriptor_s(__cs__.Structure):
        number: _c_md.uint32
        major: _c_md.uint32
        minor: _c_md.uint32
        raid_disk: _c_md.uint32
        state: _c_md.uint32
        reserved: __cs__.Array[_c_md.uint32]
        @overload
        def __init__(
            self,
            number: _c_md.uint32 | None = ...,
            major: _c_md.uint32 | None = ...,
            minor: _c_md.uint32 | None = ...,
            raid_disk: _c_md.uint32 | None = ...,
            state: _c_md.uint32 | None = ...,
            reserved: __cs__.Array[_c_md.uint32] | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    mdp_disk_t: TypeAlias = mdp_device_descriptor_s
    class mdp_superblock_s(__cs__.Structure):
        md_magic: _c_md.uint32
        major_version: _c_md.uint32
        minor_version: _c_md.uint32
        patch_version: _c_md.uint32
        gvalid_words: _c_md.uint32
        set_uuid0: __cs__.CharArray
        ctime: _c_md.uint32
        level: _c_md.uint32
        size: _c_md.uint32
        nr_disks: _c_md.uint32
        raid_disks: _c_md.uint32
        md_minor: _c_md.uint32
        not_persistent: _c_md.uint32
        set_uuid1: __cs__.CharArray
        set_uuid2: __cs__.CharArray
        set_uuid3: __cs__.CharArray
        gstate_creserved: __cs__.Array[_c_md.uint32]
        utime: _c_md.uint32
        state: _c_md.uint32
        active_disks: _c_md.uint32
        working_disks: _c_md.uint32
        failed_disks: _c_md.uint32
        spare_disks: _c_md.uint32
        sb_csum: _c_md.uint32
        events_lo: _c_md.uint32
        events_hi: _c_md.uint32
        cp_events_lo: _c_md.uint32
        cp_events_hi: _c_md.uint32
        recovery_cp: _c_md.uint32
        reshape_position: _c_md.uint64
        new_level: _c_md.uint32
        delta_disks: _c_md.uint32
        new_layout: _c_md.uint32
        new_chunk: _c_md.uint32
        gstate_sreserved: __cs__.Array[_c_md.uint32]
        layout: _c_md.uint32
        chunk_size: _c_md.uint32
        root_pv: _c_md.uint32
        root_block: _c_md.uint32
        pstate_reserved: __cs__.Array[_c_md.uint32]
        class mdp_device_descriptor_s(__cs__.Structure):
            number: _c_md.uint32
            major: _c_md.uint32
            minor: _c_md.uint32
            raid_disk: _c_md.uint32
            state: _c_md.uint32
            reserved: __cs__.Array[_c_md.uint32]
            @overload
            def __init__(
                self,
                number: _c_md.uint32 | None = ...,
                major: _c_md.uint32 | None = ...,
                minor: _c_md.uint32 | None = ...,
                raid_disk: _c_md.uint32 | None = ...,
                state: _c_md.uint32 | None = ...,
                reserved: __cs__.Array[_c_md.uint32] | None = ...,
            ): ...
            @overload
            def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

        disks: __cs__.Array[mdp_device_descriptor_s]
        reserved: __cs__.Array[_c_md.uint32]
        this_disk: _c_md.mdp_device_descriptor_s
        @overload
        def __init__(
            self,
            md_magic: _c_md.uint32 | None = ...,
            major_version: _c_md.uint32 | None = ...,
            minor_version: _c_md.uint32 | None = ...,
            patch_version: _c_md.uint32 | None = ...,
            gvalid_words: _c_md.uint32 | None = ...,
            set_uuid0: __cs__.CharArray | None = ...,
            ctime: _c_md.uint32 | None = ...,
            level: _c_md.uint32 | None = ...,
            size: _c_md.uint32 | None = ...,
            nr_disks: _c_md.uint32 | None = ...,
            raid_disks: _c_md.uint32 | None = ...,
            md_minor: _c_md.uint32 | None = ...,
            not_persistent: _c_md.uint32 | None = ...,
            set_uuid1: __cs__.CharArray | None = ...,
            set_uuid2: __cs__.CharArray | None = ...,
            set_uuid3: __cs__.CharArray | None = ...,
            gstate_creserved: __cs__.Array[_c_md.uint32] | None = ...,
            utime: _c_md.uint32 | None = ...,
            state: _c_md.uint32 | None = ...,
            active_disks: _c_md.uint32 | None = ...,
            working_disks: _c_md.uint32 | None = ...,
            failed_disks: _c_md.uint32 | None = ...,
            spare_disks: _c_md.uint32 | None = ...,
            sb_csum: _c_md.uint32 | None = ...,
            events_lo: _c_md.uint32 | None = ...,
            events_hi: _c_md.uint32 | None = ...,
            cp_events_lo: _c_md.uint32 | None = ...,
            cp_events_hi: _c_md.uint32 | None = ...,
            recovery_cp: _c_md.uint32 | None = ...,
            reshape_position: _c_md.uint64 | None = ...,
            new_level: _c_md.uint32 | None = ...,
            delta_disks: _c_md.uint32 | None = ...,
            new_layout: _c_md.uint32 | None = ...,
            new_chunk: _c_md.uint32 | None = ...,
            gstate_sreserved: __cs__.Array[_c_md.uint32] | None = ...,
            layout: _c_md.uint32 | None = ...,
            chunk_size: _c_md.uint32 | None = ...,
            root_pv: _c_md.uint32 | None = ...,
            root_block: _c_md.uint32 | None = ...,
            pstate_reserved: __cs__.Array[_c_md.uint32] | None = ...,
            disks: __cs__.Array[mdp_device_descriptor_s] | None = ...,
            reserved: __cs__.Array[_c_md.uint32] | None = ...,
            this_disk: _c_md.mdp_device_descriptor_s | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    mdp_super_t: TypeAlias = mdp_superblock_s
    class mdp_superblock_1(__cs__.Structure):
        magic: _c_md.uint32
        major_version: _c_md.uint32
        feature_map: _c_md.uint32
        pad0: _c_md.uint32
        set_uuid: __cs__.CharArray
        set_name: __cs__.CharArray
        ctime: _c_md.uint64
        level: _c_md.int32
        layout: _c_md.uint32
        size: _c_md.uint64
        chunksize: _c_md.uint32
        raid_disks: _c_md.uint32
        bitmap_offset: _c_md.uint32
        class __anonymous_0__(__cs__.Structure):
            offset: _c_md.uint16
            size: _c_md.uint16
            @overload
            def __init__(self, offset: _c_md.uint16 | None = ..., size: _c_md.uint16 | None = ...): ...
            @overload
            def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

        ppl: __anonymous_0__
        new_level: _c_md.uint32
        reshape_position: _c_md.uint64
        delta_disks: _c_md.uint32
        new_layout: _c_md.uint32
        new_chunk: _c_md.uint32
        new_offset: _c_md.uint32
        data_offset: _c_md.uint64
        data_size: _c_md.uint64
        super_offset: _c_md.uint64
        recovery_offset: _c_md.uint64
        journal_tail: _c_md.uint64
        dev_number: _c_md.uint32
        cnt_corrected_read: _c_md.uint32
        device_uuid: __cs__.CharArray
        devflags: _c_md.uint8
        bblog_shift: _c_md.uint8
        bblog_size: _c_md.uint16
        bblog_offset: _c_md.int32
        utime: _c_md.uint64
        events: _c_md.uint64
        resync_offset: _c_md.uint64
        sb_csum: _c_md.uint32
        max_dev: _c_md.uint32
        pad3: __cs__.CharArray
        dev_roles: __cs__.Array[_c_md.uint16]
        @overload
        def __init__(
            self,
            magic: _c_md.uint32 | None = ...,
            major_version: _c_md.uint32 | None = ...,
            feature_map: _c_md.uint32 | None = ...,
            pad0: _c_md.uint32 | None = ...,
            set_uuid: __cs__.CharArray | None = ...,
            set_name: __cs__.CharArray | None = ...,
            ctime: _c_md.uint64 | None = ...,
            level: _c_md.int32 | None = ...,
            layout: _c_md.uint32 | None = ...,
            size: _c_md.uint64 | None = ...,
            chunksize: _c_md.uint32 | None = ...,
            raid_disks: _c_md.uint32 | None = ...,
            bitmap_offset: _c_md.uint32 | None = ...,
            ppl: __anonymous_0__ | None = ...,
            new_level: _c_md.uint32 | None = ...,
            reshape_position: _c_md.uint64 | None = ...,
            delta_disks: _c_md.uint32 | None = ...,
            new_layout: _c_md.uint32 | None = ...,
            new_chunk: _c_md.uint32 | None = ...,
            new_offset: _c_md.uint32 | None = ...,
            data_offset: _c_md.uint64 | None = ...,
            data_size: _c_md.uint64 | None = ...,
            super_offset: _c_md.uint64 | None = ...,
            recovery_offset: _c_md.uint64 | None = ...,
            journal_tail: _c_md.uint64 | None = ...,
            dev_number: _c_md.uint32 | None = ...,
            cnt_corrected_read: _c_md.uint32 | None = ...,
            device_uuid: __cs__.CharArray | None = ...,
            devflags: _c_md.uint8 | None = ...,
            bblog_shift: _c_md.uint8 | None = ...,
            bblog_size: _c_md.uint16 | None = ...,
            bblog_offset: _c_md.int32 | None = ...,
            utime: _c_md.uint64 | None = ...,
            events: _c_md.uint64 | None = ...,
            resync_offset: _c_md.uint64 | None = ...,
            sb_csum: _c_md.uint32 | None = ...,
            max_dev: _c_md.uint32 | None = ...,
            pad3: __cs__.CharArray | None = ...,
            dev_roles: __cs__.Array[_c_md.uint16] | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

# Technically `c_md` is an instance of `_c_md`, but then we can't use it in type hints
c_md: TypeAlias = _c_md

SECTOR_SIZE: int



# This file is still a WIP
from __future__ import annotations

import logging
import os
import uuid
from collections import defaultdict
from functools import cached_property
from typing import TYPE_CHECKING, BinaryIO

from dissect.cstruct import cstruct
from dissect.util.stream import AlignedStream

from dissect.volume.exceptions import Error

if TYPE_CHECKING:
    from dissect.cstruct import Structure

log = logging.getLogger(__name__)
log.setLevel(os.getenv("DISSECT_LOG_VSS", "CRITICAL"))

vss_def = """
flag _VSS_VOLUME_SNAPSHOT_ATTRIBUTES : uint32 {
    VSS_VOLSNAP_ATTR_PERSISTENT             = 0x00000001,
    VSS_VOLSNAP_ATTR_NO_AUTORECOVERY        = 0x00000002,
    VSS_VOLSNAP_ATTR_CLIENT_ACCESSIBLE      = 0x00000004,
    VSS_VOLSNAP_ATTR_NO_AUTO_RELEASE        = 0x00000008,
    VSS_VOLSNAP_ATTR_NO_WRITERS             = 0x00000010,
    VSS_VOLSNAP_ATTR_TRANSPORTABLE          = 0x00000020,
    VSS_VOLSNAP_ATTR_NOT_SURFACED           = 0x00000040,
    VSS_VOLSNAP_ATTR_NOT_TRANSACTED         = 0x00000080,
    VSS_VOLSNAP_ATTR_HARDWARE_ASSISTED      = 0x00010000,
    VSS_VOLSNAP_ATTR_DIFFERENTIAL           = 0x00020000,
    VSS_VOLSNAP_ATTR_PLEX                   = 0x00040000,
    VSS_VOLSNAP_ATTR_IMPORTED               = 0x00080000,
    VSS_VOLSNAP_ATTR_EXPOSED_LOCALLY        = 0x00100000,
    VSS_VOLSNAP_ATTR_EXPOSED_REMOTELY       = 0x00200000,
    VSS_VOLSNAP_ATTR_AUTORECOVER            = 0x00400000,
    VSS_VOLSNAP_ATTR_ROLLBACK_RECOVERY      = 0x00800000,
    VSS_VOLSNAP_ATTR_DELAYED_POSTSNAPSHOT   = 0x01000000,
    VSS_VOLSNAP_ATTR_TXF_RECOVERY           = 0x02000000
};

enum RECORD_TYPE : uint32 {
    VOLUME_HEADER       = 0x1,
    CATALOG             = 0x2,
    STORE_INDEX         = 0x3,
    STORE_HEADER        = 0x4,
    STORE_BLOCK_RANGE   = 0x5,
    STORE_BITMAP        = 0x6
};

flag BLOCK_FLAG : uint32 {
    IS_FORWARDER    = 0x1,
    IS_OVERLAY      = 0x2,
    NOT_USED        = 0x4
};

struct volume_header {
    char    identifier[16];
    uint32  version;
    uint32  record_type;
    uint64  current_offset;
    uint64  unk0;
    uint64  unk1;
    uint64  catalog_offset;
    uint64  maximum_size;
    char    volume_identifier[16];
    char    store_volume_identifier[16];
    uint32  unk2;
    char    unk3[412];
};

struct catalog_header {
    char    identifier[16];
    uint32  version;
    uint32  record_type;
    uint64  relative_offset;
    uint64  offset;
    uint64  next_offset;
    char    unk0[80];
};

struct catalog_entry_1 {
    uint64  entry_type;
    char    unk0[120];
};

struct catalog_entry_2 {
    uint64  entry_type;
    uint64  volume_size;
    char    store_identifier[16];
    uint64  unk0;
    uint64  unk1;
    uint64  creation_time;
    char    unk2[72];
};

struct catalog_entry_3 {
    uint64  entry_type;
    uint64  store_block_list_offset;
    char    store_identifier[16];
    uint64  store_header_offset;
    uint64  store_range_list_offset;
    uint64  store_bitmap_offset;
    uint64  metadata_reference;
    uint64  allocated_size;
    uint64  store_previous_bitmap_offset;
    uint64  unk0;
    char    unk1[40];
};

struct store_header {
    char        identifier[16];
    uint32      version;
    RECORD_TYPE record_type;
    uint64      relative_offset;
    uint64      offset;
    uint64      next_offset;
    uint64      size;
    char        unk0[72];
};

struct store_information {
    char    unk_identifier[16];
    char    copy_identifier[16];
    char    copy_set_identifier[16];
    uint32  type;
    uint32  provider;
    _VSS_VOLUME_SNAPSHOT_ATTRIBUTES attributes;
    uint32  unk0;
    uint16  operating_machine_len;
    wchar   operating_machine[operating_machine_len / 2];
    uint16  service_machine_len;
    wchar   service_machine[service_machine_len / 2];
};

struct block_descriptor {
    uint64  original_offset;
    uint64  relative_offset;
    uint64  store_offset;
    BLOCK_FLAG  flags;
    uint32  allocation_bitmap;
};

struct range_descriptor {
    uint64  store_offset;
    uint64  relative_offset;
    uint64  size;
};
"""

c_vss = cstruct().load(vss_def)

RECORD_TYPE = c_vss.RECORD_TYPE
BLOCK_FLAG = c_vss.BLOCK_FLAG

VSS_IDENTIFIER = b"\x6b\x87\x08\x38\x76\xc1\x48\x4e\xb7\xae\x04\x04\x6e\x6c\xc7\x52"
VOLUME_HEADER_OFFSET = 0x1E00
BLOCK_SIZE = 0x4000

CATALOG_BLOCK_SIZE = 0x4000
CATALOG_ENTRY_SIZE = 128

STORE_BLOCK_SIZE = 0x4000

STORE_BLOCKLIST_ENTRY_SIZE = 32
STORE_RANGELIST_ENTRY_SIZE = 24


class VSS:
    def __init__(self, fh: BinaryIO):
        self.fh = fh

        fh.seek(VOLUME_HEADER_OFFSET)
        self.header = c_vss.volume_header(fh)
        if self.header.identifier != VSS_IDENTIFIER:
            raise Error(
                f"Not a valid VSS identifier (got {self.header.identifier.hex()}, expected {VSS_IDENTIFIER.hex()}"
            )

        if self.header.catalog_offset == 0:
            raise Error("Catalog offset is 0")

        self.catalog = Catalog(self, self.header.catalog_offset)

    def __repr__(self) -> str:
        return (
            f"<VSS volume_identifier={self.volume_identifier} store_volume_identifier={self.store_volume_identifier}>"
        )

    @property
    def volume_identifier(self) -> uuid.UUID:
        return uuid.UUID(bytes_le=self.header.volume_identifier)

    @property
    def store_volume_identifier(self) -> uuid.UUID:
        return uuid.UUID(bytes_le=self.header.store_volume_identifier)


class Catalog:
    def __init__(self, vss: VSS, offset: int):
        self.vss = vss
        self.fh = vss.fh
        self.header, data = read_block(vss.fh, offset, c_vss.catalog_header)
        buf = memoryview(data)

        store_map = defaultdict(list)

        self.entries = []
        for entry_offset in range(0, len(data), CATALOG_ENTRY_SIZE):
            entry_buf = buf[entry_offset : entry_offset + CATALOG_ENTRY_SIZE]
            entry_type = c_vss.uint64(entry_buf[:8])

            entry = None
            if entry_type == 1:
                entry = c_vss.catalog_entry_1(entry_buf)
            elif entry_type == 2:
                entry = c_vss.catalog_entry_2(entry_buf)
            elif entry_type == 3:
                entry = c_vss.catalog_entry_3(entry_buf)
            elif entry_type == 0:
                break
            else:
                raise Error("Invalid catalog entry")

            if entry_type in (2, 3):
                store_map[entry.store_identifier].append(entry)

            self.entries.append(entry)

        self.stores = []
        for store_descriptors in store_map.values():
            self.stores.append(Store(self, store_descriptors))

        self.stores = sorted(self.stores, key=lambda store: store.creation_time)
        prev_store = None
        for i, store in enumerate(self.stores):
            store.index = i
            store.previous_store = prev_store
            if prev_store:
                prev_store.next_store = store
            prev_store = store

    def __repr__(self) -> str:
        return f"<Catalog stores={len(self.stores)}>"


DEBUG = False


class Store:
    def __init__(self, catalog: Catalog, descriptors: list):
        self.catalog = catalog
        self.fh = catalog.fh
        self.descriptors = descriptors
        self.previous_store = None
        self.next_store = None
        self.index = None

        if len(descriptors) != 2:
            raise NotImplementedError("Only got one descriptor")

        self.has_store_data = False
        if len(descriptors) == 2:
            self.has_store_data = True

        self.desc2, self.desc3 = descriptors
        self.volume_size = self.desc2.volume_size
        self.size = self.volume_size
        self.creation_time = self.desc2.creation_time

        # Only if store data
        self.fh.seek(self.desc3.store_header_offset)
        self.header = c_vss.store_header(self.fh)

        if self.header.identifier != VSS_IDENTIFIER:
            raise Error(
                f"Not a valid VSS identifier (got {self.header.identifier.hex()}, expected {VSS_IDENTIFIER.hex()}"
            )

        self.information = c_vss.store_information(self.fh)

        self.copy_identifier = uuid.UUID(bytes_le=self.information.copy_identifier)
        self.copy_set_identifier = uuid.UUID(bytes_le=self.information.copy_set_identifier)

        self._block_list = None
        self._range_list = None
        self._bitmap = None
        self._previous_bitmap = None

    def open(self) -> StoreStream:
        return StoreStream(self)

    @cached_property
    def block_list(self) -> BlockList:
        return BlockList(self, self.desc3.store_block_list_offset)

    @cached_property
    def range_list(self) -> RangeList:
        return RangeList(self, self.desc3.store_range_list_offset)

    @cached_property
    def bitmap(self) -> StoreBitmap:
        return StoreBitmap(self, self.desc3.store_bitmap_offset)

    @cached_property
    def previous_bitmap(self) -> StoreBitmap:
        return StoreBitmap(self, self.desc3.store_previous_bitmap_offset)

    def read_block(self, block: int, active_store: Store | None = None) -> bytes:
        active_store = active_store or self

        buf = None
        descriptor = self.block_list.map.map.get(block)
        if DEBUG:
            print(block, descriptor)
        # if descriptor:
        #     import ipdb; ipdb.set_trace()

        # if descriptor and descriptor.is_overlay:
        #     import ipdb; ipdb.set_trace()

        if descriptor:
            if descriptor.is_forwarder:
                if self.next_store:
                    print("DEBUG THIS")
                    buf = self.next_store.read_block(descriptor.relative_offset // 0x4000, self)
                else:
                    self.fh.seek(descriptor.relative_offset)
                    buf = self.fh.read(BLOCK_SIZE)
            elif not descriptor.is_overlay:
                # print("reading from store")
                self.fh.seek(descriptor.store_offset)
                buf = self.fh.read(BLOCK_SIZE)

        if not descriptor or descriptor.is_overlay:
            if self.next_store:
                # print("reading from next store")
                buf = self.next_store.read_block(block, self)
            else:
                reverse_descriptor = self.block_list.map.reverse.get(block)
                if reverse_descriptor:
                    # print("found reverse descriptor, reading from volume")
                    self.fh.seek(block * BLOCK_SIZE)
                    buf = self.fh.read(BLOCK_SIZE)
                elif self.bitmap.in_use(block) and (not self.previous_bitmap or self.previous_bitmap.in_use(block)):
                    # print("sparse block?")
                    buf = b"\x00" * BLOCK_SIZE
                else:
                    # print("nothing matched, reading from volume")
                    self.fh.seek(block * BLOCK_SIZE)
                    buf = self.fh.read(BLOCK_SIZE)

        if not buf:
            print("booboo")
            print("Descriptor", descriptor)
            raise ValueError("Error reading block")

        if descriptor and active_store is self and (descriptor.is_overlay or descriptor.overlay):
            # print("overlaying data")
            buf = bytearray(buf)
            overlay = descriptor if descriptor.is_overlay else descriptor.overlay
            overlay_offset = overlay.store_offset
            bitmap = overlay.bitmap

            for i in range(32):
                if (bitmap >> i) & 0x00000001:
                    self.fh.seek(overlay_offset + (i * 512))
                    buf[i * 512 : (i + 1) * 512] = self.fh.read(512)

        return bytes(buf)

        #     if descriptor.is_forwarder:
        #         block_data_offset = descriptor.relative_offset
        #     else:
        #         block_data_offset = descriptor.store_offset

        #     if overlay:
        #         if active_store is not self and descriptor is overlay:
        #             print('weird shit')
        #             descriptor = None
        #         else:
        #             overlay_offset = overlay.original_offset
        #             overlay_bitmap = overlay.bitmap
        #             pass

        # if descriptor:
        #     if descriptor.is_forwarder:
        #         block_data_offset = descriptor.relative_offset
        #     else:
        #         block_data_offset = descriptor.store_offset

        #     if descriptor.is_overlay:
        #         overlay_block_descriptor = descriptor
        #     else:
        #         overlay_block_descriptor = descriptor.overlay

        #     if overlay_block_descriptor:
        #         if self.index != active_store.index and descriptor is overlay_block_descriptor:
        #             descriptor = None
        #         else:
        #             overlay_block_offset = overlay_block_descriptor.original_offset
        #             overlay_bitmap = overlay_block_descriptor.bitmap
        #             # bitmap stuff

        # if not descriptor:
        #     if not self.next_store and self.index == active_store.index:
        #         reverse_block_descriptor = self.block_list.map.reverse.get(block)
        #         in_bitmap = self.bitmap.in_use(block)
        #         in_previous_bitmap = self.previous_bitmap.in_use(block)

        # if descriptor:
        #     if descriptor.is_forwarder and self.next_store:
        #         print("reading forwarder from next store")
        #         return self.next_store.read_block(block, active_store)
        #     else:
        #         print("reading from current store")
        #         self.fh.seek(block_data_offset)
        #         return self.fh.read(BLOCK_SIZE)
        # else:
        #     if self.next_store:
        #         print("reading from next store")
        #         return self.next_store.read_block(block, active_store)
        #     elif not reverse_block_descriptor and not in_bitmap and not in_previous_bitmap:
        #         print("zero'ing")
        #         return b'\x00' * BLOCK_SIZE
        #     else:
        #         print("reading from volume")
        #         self.fh.seek(block * BLOCK_SIZE)
        #         return self.fh.read(BLOCK_SIZE)


class StoreStream(AlignedStream):
    def __init__(self, store: Store):
        self.store = store
        super().__init__(size=store.size, align=BLOCK_SIZE)

    def _read(self, offset: int, length: int) -> bytes:
        r = []
        blockidx = offset // BLOCK_SIZE

        while length > 0:
            r.append(self.store.read_block(blockidx))
            blockidx += 1
            length -= BLOCK_SIZE

        return b"".join(r)


class BlockList:
    def __init__(self, store: Store, offset: int):
        self.store = store
        self.offset = offset

        # t = time.time()
        self.header, data = read_block(store.fh, offset, c_vss.store_header)
        # print time.time() - t

        if self.header.record_type != RECORD_TYPE.STORE_INDEX:
            raise Error(
                "invalid store header type for block list"
                f" (got 0x{self.header.record_type:x}, expected 0x{RECORD_TYPE.STORE_INDEX:x}"
            )

        # t = time.time()
        buf = memoryview(data)
        self.map = BlockMap()
        for entry_offset in range(0, len(data), STORE_BLOCKLIST_ENTRY_SIZE):
            entry_buf = buf[entry_offset : entry_offset + STORE_BLOCKLIST_ENTRY_SIZE]
            if entry_buf == b"\x00" * STORE_BLOCKLIST_ENTRY_SIZE:
                break
            entry = BlockDescriptor(entry_buf)
            # self.entries.append(entry)
            self.map.add(entry)
        # print time.time() - t


class RangeList:
    def __init__(self, store: Store, offset: int):
        self.store = store
        self.offset = offset

        self.header, data = read_block(store.fh, offset, c_vss.store_header)

        if self.header.record_type != RECORD_TYPE.STORE_BLOCK_RANGE:
            raise Error(
                "invalid store header type for block list"
                f" (got 0x{self.header.record_type:x}, expected 0x{RECORD_TYPE.STORE_BLOCK_RANGE:x}"
            )

        buf = memoryview(data)
        self.entries = []
        for entry_offset in range(0, len(data), STORE_RANGELIST_ENTRY_SIZE):
            entry_buf = buf[entry_offset : entry_offset + STORE_RANGELIST_ENTRY_SIZE]
            if entry_buf == b"\x00" * STORE_RANGELIST_ENTRY_SIZE:
                break
            entry = c_vss.range_descriptor(entry_buf)
            self.entries.append(entry)


class StoreBitmap:
    def __init__(self, store: Store, offset: int):
        self.store = store
        self.offset = offset

        self.header, self.data = read_block(store.fh, offset, c_vss.store_header)

        if self.header.record_type != RECORD_TYPE.STORE_BITMAP:
            raise Error(
                "invalid store header type for bitmap"
                f" (got 0x{self.header.record_type:x}, expected 0x{RECORD_TYPE.STORE_BITMAP:x}"
            )

        buf = memoryview(self.data)
        self.test = []
        for i in range(0, len(self.data) // 4, 4):
            b = buf[i : i + 4]
            val = c_vss.uint32(b)
            for _ in range(32):
                if val & 0x00000001 == 0:
                    self.test.append(False)
                else:
                    self.test.append(True)
                val >>= 1

    def has_offset(self, offset: int) -> bool:
        return self.in_use(offset // BLOCK_SIZE)

    def in_use(self, block: int) -> bool:
        return not self.is_set(block)

    def is_set(self, block: int) -> bool:
        return (self.data[block // 8] & (1 << (block % 8))) != 0

    def __getitem__(self, block: int) -> bool:
        return self.in_use(block)


class BlockMap:
    def __init__(self):
        self.map = {}
        self.reverse = {}

    def add(self, descriptor: BlockDescriptor) -> None:
        if not descriptor.is_used:
            return

        blockmap = self.map
        reversemap = self.reverse
        key = descriptor.original_offset // BLOCK_SIZE

        if not descriptor.is_overlay:
            revkey = descriptor.relative_offset // BLOCK_SIZE
            try:
                revexist = reversemap[revkey]
            except KeyError:
                revexist = None

            if revexist:
                descriptor.original_offset = revexist.relative_offset
                del reversemap[revkey]

        if descriptor.is_forwarder and descriptor.original_offset == descriptor.relative_offset:
            return

        try:
            existing = blockmap[key]
        except KeyError:
            existing = None

        if existing:
            if descriptor.is_overlay:
                overlay = existing if existing.is_overlay else existing.overlay

                if overlay:
                    overlay.bitmap |= descriptor.bitmap
                else:
                    existing.overlay = descriptor
                return

            if existing.is_overlay:
                descriptor.overlay = existing
            else:
                descriptor.overlay = existing.overlay

        blockmap[key] = descriptor

        if descriptor.is_forwarder:
            revkey = descriptor.relative_offset // BLOCK_SIZE
            reversemap[revkey] = descriptor

    def get_descriptor(self, offset: int) -> None:
        pass

    def __getitem__(self, block: int) -> BlockMap:
        return self


class BlockDescriptor:
    __slots__ = (
        "bitmap",
        "flags",
        "is_forwarder",
        "is_overlay",
        "is_used",
        "original_offset",
        "overlay",
        "relative_offset",
        "store",
        "store_offset",
    )

    def __init__(self, buf: bytes):
        # self.store = store

        entry = c_vss.block_descriptor(buf)
        self.original_offset = entry.original_offset
        self.relative_offset = entry.relative_offset
        self.store_offset = entry.store_offset
        self.flags = entry.flags
        self.bitmap = entry.allocation_bitmap

        self.overlay = None
        self.is_used = not bool(self.flags & BLOCK_FLAG.NOT_USED)
        self.is_overlay = bool(self.flags & BLOCK_FLAG.IS_OVERLAY)
        self.is_forwarder = bool(self.flags & BLOCK_FLAG.IS_FORWARDER)

    def __eq__(self, other: object) -> bool:
        if not isinstance(BlockDescriptor, other):
            return False

        return (
            self.original_offset == other.original_offset
            and self.relative_offset == other.relative_offset
            and self.store_offset == other.store_offset
            and self.flags == other.flags
            and self.bitmap == other.bitmap
        )

    def __repr__(self) -> str:
        return (
            f"<BlockDescriptor original_offset=0x{self.original_offset:08x}"
            f" relative_offset=0x{self.relative_offset:08x} store_offset=0x{self.store_offset:08x}"
            f" flags={self.flags} bitmap=0x{self.bitmap:08x}>"
        )


# libvshadow_store_descriptor_read_buffer


def read_block(fh: BinaryIO, offset: int, struct: type[Structure]) -> tuple[Structure, bytes]:
    header, buf = read_block_data(fh, offset, struct)

    r = [buf]
    offset = header.next_offset
    while offset != 0:
        nheader, buf = read_block_data(fh, offset, struct)
        offset = nheader.next_offset
        r.append(buf)

    return header, b"".join(r)


def read_block_data(fh: BinaryIO, offset: int, struct: type[Structure]) -> tuple[Structure, bytes]:
    fh.seek(offset)
    buf = fh.read(BLOCK_SIZE)
    header = struct(buf)

    if header.identifier != VSS_IDENTIFIER:
        raise Error(f"not a valid VSS identifier (got {header.identifier.hex()}, expected {VSS_IDENTIFIER.hex()}")

    return header, buf[128:]



from __future__ import annotations

from typing import TYPE_CHECKING, BinaryIO
from uuid import UUID

from dissect.util.stream import RangeStream

if TYPE_CHECKING:
    from dissect.cstruct import Structure

    from dissect.volume.disk.schemes import APM, BSD, GPT, MBR


PARTITION_TYPES = {
    0x00: "Empty",
    0x01: "FAT12,CHS",
    0x04: "FAT16 16-32MB,CHS",
    0x05: "Microsoft Extended",
    0x06: "FAT16 32MB,CHS",
    0x07: "NTFS",
    0x0B: "FAT32,CHS",
    0x0C: "FAT32,LBA",
    0x0E: "FAT16, 32MB-2GB,LBA",
    0x0F: "Microsoft Extended, LBA",
    0x11: "Hidden FAT12,CHS",
    0x14: "Hidden FAT16,16-32MB,CHS",
    0x16: "Hidden FAT16,32MB-2GB,CHS",
    0x18: "AST SmartSleep Partition",
    0x1B: "Hidden FAT32,CHS",
    0x1C: "Hidden FAT32,LBA",
    0x1E: "Hidden FAT16,32MB-2GB,LBA",
    0x27: "PQservice",
    0x39: "Plan 9 partition",
    0x3C: "PartitionMagic recovery partition",
    0x42: "Microsoft MBR,Dynamic Disk",
    0x44: "GoBack partition",
    0x51: "Novell",
    0x52: "CP/M",
    0x63: "Unix System V",
    0x64: "PC-ARMOUR protected partition",
    0x82: "Solaris x86 or Linux Swap",
    0x83: "Linux",
    0x84: "Hibernation",
    0x85: "Linux Extended",
    0x86: "NTFS Volume Set",
    0x87: "NTFS Volume Set",
    0x9F: "BSD/OS",
    0xA0: "Hibernation",
    0xA1: "Hibernation",
    0xA5: "FreeBSD",
    0xA6: "OpenBSD",
    0xA8: "Mac OSX",
    0xA9: "NetBSD",
    0xAB: "Mac OSX Boot",
    0xAF: "MacOS X HFS",
    0xB7: "BSDI",
    0xB8: "BSDI Swap",
    0xBB: "Boot Wizard hidden",
    0xBE: "Solaris 8 boot partition",
    0xD8: "CP/M-86",
    0xDE: "Dell PowerEdge Server utilities (FAT fs)",
    0xDF: "DG/UX virtual disk manager partition",
    0xEB: "BeOS BFS",
    0xEE: "EFI GPT Disk",
    0xEF: "EFI System Partition",
    0xFB: "VMWare File System",
    0xFC: "VMWare Swap",
    # https://gist.github.com/Alex131089/b3a23c9461e95433387f285f6e0860ca
    UUID("00000000-0000-0000-0000-000000000000"): "Unused entry",
    UUID("024DEE41-33E7-11D3-9D69-0008C781F39F"): "MBR partition scheme",
    UUID("C12A7328-F81F-11D2-BA4B-00A0C93EC93B"): "EFI System partition",
    UUID("21686148-6449-6E6F-744E-656564454649"): "BIOS Boot partition",
    UUID("D3BFE2DE-3DAF-11DF-BA40-E3A556D89593"): (
        "Intel Fast Flash (iFFS) partition (for Intel Rapid Start technology)"
    ),
    UUID("F4019732-066E-4E12-8273-346C5641494F"): "Sony boot partition",
    UUID("BFBFAFE7-A34F-448A-9A5B-6213EB736C22"): "Lenovo boot partition",
    UUID("E3C9E316-0B5C-4DB8-817D-F92DF00215AE"): "Microsoft Reserved Partition (MSR)",
    UUID("EBD0A0A2-B9E5-4433-87C0-68B6B72699C7"): "Windows Basic data partition",
    UUID("5808C8AA-7E8F-42E0-85D2-E1E90434CFB3"): "Windows Logical Disk Manager (LDM) metadata",
    UUID("AF9B60A0-1431-4F62-BC68-3311714A69AD"): "Windows Logical Disk Manager data",
    UUID("DE94BBA4-06D1-4D40-A16A-BFD50179D6AC"): "Windows Recovery Environment",
    UUID("37AFFC90-EF7D-4E96-91C3-2D7AE055B174"): "Windows IBM General Parallel File System (GPFS)",
    UUID("E75CAF8F-F680-4CEE-AFA3-B001E56EFC2D"): "Windows Storage Spaces partition",
    UUID("75894C1E-3AEB-11D3-B7C1-7B03A0000000"): "HP-UX Data",
    UUID("E2A1E728-32E3-11D6-A682-7B03A0000000"): "HP-UX Service",
    UUID("0FC63DAF-8483-4772-8E79-3D69D8477DE4"): "Linux filesystem data",
    UUID("A19D880F-05FC-4D3B-A006-743F0F84911E"): "Linux RAID",
    UUID("44479540-F297-41B2-9AF7-D131D5F0458A"): "Linux Root (x86)",
    UUID("4F68BCE3-E8CD-4DB1-96E7-FBCAF984B709"): "Linux Root (x86-64)",
    UUID("69DAD710-2CE4-4E3C-B16C-21A1D49ABED3"): "Linux Root (32-bit ARM)",
    UUID("B921B045-1DF0-41C3-AF44-4C6F280D3FAE"): "Linux Root (64-bit ARM/AArch64)",
    UUID("993D8D3D-F80E-4225-855A-9DAF8ED7EA97"): "Linux Root (IA-64)",
    UUID("0657FD6D-A4AB-43C4-84E5-0933C84B4F4F"): "Linux Swap",
    UUID("E6D6D379-F507-44C2-A23C-238F2A3DF928"): "Linux Logical Volume Manager (LVM)",
    UUID("933AC7E1-2EB4-4F13-B844-0E14E2AEF915"): "Linux /home",
    UUID("3B8F8425-20E0-4F3B-907F-1A25A76F98E8"): "Linux /srv (server data)",
    UUID("7FFEC5C9-2D00-49B7-8941-3EA10A5586B7"): "Linux Plain dm-crypt",
    UUID("CA7D7CCB-63ED-4C53-861C-1742536059CC"): "Linux LUKS",
    UUID("8DA63339-0007-60C0-C436-083AC8230908"): "Linux Reserved",
    UUID("83BD6B9D-7F41-11DC-BE0B-001560B84F0F"): "FreeBSD Boot",
    UUID("516E7CB4-6ECF-11D6-8FF8-00022D09712B"): "FreeBSD Data",
    UUID("516E7CB5-6ECF-11D6-8FF8-00022D09712B"): "FreeBSD Swap",
    UUID("516E7CB6-6ECF-11D6-8FF8-00022D09712B"): "FreeBSD Unix File System (UFS)",
    UUID("516E7CB8-6ECF-11D6-8FF8-00022D09712B"): "FreeBSD Vinum volume manager",
    UUID("516E7CBA-6ECF-11D6-8FF8-00022D09712B"): "FreeBSD ZFS",
    UUID("9D087404-1CA5-11DC-8817-01301BB8A9F5"): "DragonFlyBSD disklabel32",
    UUID("3D48CE54-1D16-11DC-8696-01301BB8A9F5"): "DragonFlyBSD disklabel64",
    UUID("48465300-0000-11AA-AA11-00306543ECAC"): "Apple Hierarchical File System Plus (HFS+)",
    UUID("55465300-0000-11AA-AA11-00306543ECAC"): "Apple UFS",
    UUID("52414944-0000-11AA-AA11-00306543ECAC"): "Apple RAID",
    UUID("52414944-5F4F-11AA-AA11-00306543ECAC"): "Apple RAID (offline)",
    UUID("426F6F74-0000-11AA-AA11-00306543ECAC"): "Apple Boot (Recovery HD)",
    UUID("4C616265-6C00-11AA-AA11-00306543ECAC"): "Apple Label",
    UUID("5265636F-7665-11AA-AA11-00306543ECAC"): "Apple TV Recovery",
    UUID("53746F72-6167-11AA-AA11-00306543ECAC"): "Apple Core Storage (i.e. Lion FileVault)",
    UUID("7C3457EF-0000-11AA-AA11-00306543ECAC"): "Apple APFS",
    UUID("6A82CB45-1DD2-11B2-99A6-080020736631"): "Solaris Boot",
    UUID("6A85CF4D-1DD2-11B2-99A6-080020736631"): "Solaris Root",
    UUID("6A87C46F-1DD2-11B2-99A6-080020736631"): "Solaris Swap",
    UUID("6A8B642B-1DD2-11B2-99A6-080020736631"): "Solaris Backup",
    UUID("6A898CC3-1DD2-11B2-99A6-080020736631"): "Solaris /usr // Apple ZFS",
    UUID("6A8EF2E9-1DD2-11B2-99A6-080020736631"): "Solaris /var",
    UUID("6A90BA39-1DD2-11B2-99A6-080020736631"): "Solaris /home",
    UUID("6A9283A5-1DD2-11B2-99A6-080020736631"): "Solaris Alternate sector",
    UUID("6A945A3B-1DD2-11B2-99A6-080020736631"): "Solaris Reserved",
    UUID("6A9630D1-1DD2-11B2-99A6-080020736631"): "Solaris Reserved",
    UUID("6A980767-1DD2-11B2-99A6-080020736631"): "Solaris Reserved",
    UUID("6A96237F-1DD2-11B2-99A6-080020736631"): "Solaris Reserved",
    UUID("6A8D2AC7-1DD2-11B2-99A6-080020736631"): "Solaris Reserved",
    UUID("49F48D32-B10E-11DC-B99B-0019D1879648"): "NetBSD Swap",
    UUID("49F48D5A-B10E-11DC-B99B-0019D1879648"): "NetBSD FFS",
    UUID("49F48D82-B10E-11DC-B99B-0019D1879648"): "NetBSD LFS",
    UUID("49F48DAA-B10E-11DC-B99B-0019D1879648"): "NetBSD RAID",
    UUID("2DB519C4-B10F-11DC-B99B-0019D1879648"): "NetBSD Concatenated partition",  # libfdisk is wrong on this
    UUID("2DB519EC-B10F-11DC-B99B-0019D1879648"): "NetBSD Encrypted partition",  # libfdisk is wrong on this
    UUID("FE3A2A5D-4F32-41A7-B725-ACCC3285A309"): "ChromeOS kernel",
    UUID("3CB8E202-3B7E-47DD-8A3C-7FF2A13CFCEC"): "ChromeOS rootfs",
    UUID("2E0A753D-9E48-43B0-8337-B15192CB1B5E"): "ChromeOS future use",
    UUID("42465331-3BA3-10F1-802A-4861696B7521"): "Haiku BFS",
    UUID("85D5E45E-237C-11E1-B4B3-E89A8F7FC3A7"): "MidnightBSD Boot",
    UUID("85D5E45A-237C-11E1-B4B3-E89A8F7FC3A7"): "MidnightBSD Data",
    UUID("85D5E45B-237C-11E1-B4B3-E89A8F7FC3A7"): "MidnightBSD Swap",
    UUID("0394EF8B-237E-11E1-B4B3-E89A8F7FC3A7"): "MidnightBSD Unix File System (UFS)",
    UUID("85D5E45C-237C-11E1-B4B3-E89A8F7FC3A7"): "MidnightBSD Vinum volume manager",
    UUID("85D5E45D-237C-11E1-B4B3-E89A8F7FC3A7"): "MidnightBSD ZFS",
    UUID("45B0969E-9B03-4F30-B4C6-B4B80CEFF106"): "Ceph Journal",
    UUID("45B0969E-9B03-4F30-B4C6-5EC00CEFF106"): "Ceph dm-crypt Encrypted Journal",
    UUID("4FBD7E29-9D25-41B8-AFD0-062C0CEFF05D"): "Ceph OSD",
    UUID("4FBD7E29-9D25-41B8-AFD0-5EC00CEFF05D"): "Ceph dm-crypt OSD",
    UUID("89C57F98-2FE5-4DC0-89C1-F3AD0CEFF2BE"): "Ceph disk in creation",
    UUID("89C57F98-2FE5-4DC0-89C1-5EC00CEFF2BE"): "Ceph dm-crypt disk in creation",
    UUID("824CC7A0-36A8-11E3-890A-952519AD3F61"): "OpenBSD Data partition",
    UUID("CEF5A9AD-73BC-4601-89F3-CDEEEEE321A1"): "QNX Power-safe (QNX6) file system",
    UUID("C91818F9-8025-47AF-89D2-F030D7000C2C"): "Plan 9 partition",
    UUID("9D275380-40AD-11DB-BF97-000C2911D1B8"): "VMware ESX vmkcore (coredump)",
    UUID("AA31E02A-400F-11DB-9590-000C2911D1B8"): "VMware ESX VMFS filesystem",
    UUID("9198EFFC-31C0-11DB-8F78-000C2911D1B8"): "VMware ESX VMware Reserved",
    UUID("2568845D-2332-4675-BC39-8FA5A4748D15"): "Android-IA Bootloader",
    UUID("114EAFFE-1552-4022-B26E-9B053604CF84"): "Android-IA Bootloader2",
    UUID("49A4D17F-93A3-45C1-A0DE-F50B2EBE2599"): "Android-IA Boot",
    UUID("4177C722-9E92-4AAB-8644-43502BFD5506"): "Android-IA Recovery",
    UUID("EF32A33B-A409-486C-9141-9FFB711F6266"): "Android-IA Misc",
    UUID("20AC26BE-20B7-11E3-84C5-6CFDB94711E9"): "Android-IA Metadata",
    UUID("38F428E6-D326-425D-9140-6E0EA133647C"): "Android-IA System",
    UUID("A893EF21-E428-470A-9E55-0668FD91A2D9"): "Android-IA Cache",
    UUID("DC76DDA9-5AC1-491C-AF42-A82591580C0D"): "Android-IA Data",
    UUID("EBC597D0-2053-4B15-8B64-E0AAC75F4DB1"): "Android-IA Persistent",
    UUID("C5A0AEEC-13EA-11E5-A1B1-001E67CA0C3C"): "Android-IA Vendor",
    UUID("BD59408B-4514-490D-BF12-9878D963F378"): "Android-IA Config",
    UUID("8F68CC74-C5E5-48DA-BE91-A0C8C15E9C80"): "Android-IA Factory",
    UUID("9FDAA6EF-4B3F-40D2-BA8D-BFF16BFB887B"): "Android-IA Factory (alt)",
    UUID("767941D0-2085-11E3-AD3B-6CFDB94711E9"): "Android-IA Fastboot / Tertiary",
    UUID("AC6D7924-EB71-4DF8-B48D-E267B27148FF"): "Android-IA OEM",
    # Wiki DE
    UUID("7412F7D5-A156-4B13-81DC-867174929325"): "ONIE boot",
    UUID("D4E6E2CD-4469-46F3-B5CB-1BFF57AFC149"): "ONIE config",
    UUID("9E1A2D38-C612-4316-AA26-8B49521E5A8B"): "PowerPC PReP boot",
    UUID("BC13C2FF-59E6-4262-A352-B275FD6F7172"): "Freedesktop Extended Boot Partition",
}


class Partition:
    """A partition on a disk.

    Args:
        disk: The disk the partition is on.
        number: The partition number (1-indexed).
        offset: The offset of the partition in bytes.
        size: The size of the partition in bytes.
        type: The partition type. This is either an integer (MBR, BSD), a string (APM),
              or a UUID object (GPT).
        name: The partition name (if available).
        flags: The partition flags (if available).
        guid: The partition GUID (if available).
        type_str: The partition type as a string (if available).
        raw: The raw partition structure (if available).
    """

    def __init__(
        self,
        disk: APM | BSD | GPT | MBR,
        number: int,
        offset: int,
        size: int,
        type: int | str | UUID,
        name: str | None = None,
        flags: int | None = None,
        guid: UUID | None = None,
        type_str: str | None = None,
        raw: Structure = None,
    ):
        self.disk = disk
        self.number = number
        self.offset = offset
        self.size = size
        self.type = type
        self.name = name
        self.flags = flags
        self.guid = guid
        self.type_str = type_str or translate_partition_type(self.type)
        self.type_name = translate_partition_type(self.type, False)
        self.raw = raw

    def __repr__(self) -> str:
        return (
            f"<Partition number={self.number} offset=0x{self.offset:x} "
            f"size=0x{self.size:x} type={self.type_str!r} name={self.name!r}>"
        )

    def open(self) -> BinaryIO:
        """Open a stream to the partition."""
        return RangeStream(self.disk.fh, offset=self.offset, size=self.size)


def translate_partition_type(type: int | str | UUID, prepend_type: bool = True) -> str:
    """Translate a partition type to a human-readable format."""
    if not isinstance(type, str):
        name = PARTITION_TYPES.get(type, "Unknown")
        type = hex(type) if isinstance(type, int) else type
        return f"{type} ({name})" if prepend_type else name

    return type



# Generated by cstruct-stubgen
from typing import BinaryIO, Literal, TypeAlias, overload

import dissect.cstruct as __cs__

class _c_dm(__cs__.cstruct):
    THIN_SUPERBLOCK_MAGIC: Literal[27022010] = ...
    THIN_SUPERBLOCK_LOCATION: Literal[0] = ...
    THIN_VERSION: Literal[2] = ...
    SECTOR_TO_BLOCK_SHIFT: Literal[3] = ...
    SPACE_MAP_ROOT_SIZE: Literal[128] = ...
    class node_flags(__cs__.Enum):
        INTERNAL_NODE = ...
        LEAF_NODE = ...

    class node_header(__cs__.Structure):
        csum: _c_dm.uint32
        flags: _c_dm.uint32
        blocknr: _c_dm.uint64
        nr_entries: _c_dm.uint32
        max_entries: _c_dm.uint32
        value_size: _c_dm.uint32
        padding: _c_dm.uint32
        @overload
        def __init__(
            self,
            csum: _c_dm.uint32 | None = ...,
            flags: _c_dm.uint32 | None = ...,
            blocknr: _c_dm.uint64 | None = ...,
            nr_entries: _c_dm.uint32 | None = ...,
            max_entries: _c_dm.uint32 | None = ...,
            value_size: _c_dm.uint32 | None = ...,
            padding: _c_dm.uint32 | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    class btree_node(__cs__.Structure):
        header: _c_dm.node_header
        @overload
        def __init__(self, header: _c_dm.node_header | None = ...): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    class thin_disk_superblock(__cs__.Structure):
        csum: _c_dm.uint32
        flags: _c_dm.uint32
        blocknr: _c_dm.uint64
        uuid: __cs__.Array[_c_dm.uint8]
        magic: _c_dm.uint64
        version: _c_dm.uint32
        time: _c_dm.uint32
        trans_id: _c_dm.uint64
        held_root: _c_dm.uint64
        data_space_map_root: __cs__.Array[_c_dm.uint8]
        metadata_space_map_root: __cs__.Array[_c_dm.uint8]
        data_mapping_root: _c_dm.uint64
        device_details_root: _c_dm.uint64
        data_block_size: _c_dm.uint32
        metadata_block_size: _c_dm.uint32
        metadata_nr_blocks: _c_dm.uint64
        compat_flags: _c_dm.uint32
        compat_ro_flags: _c_dm.uint32
        incompat_flags: _c_dm.uint32
        @overload
        def __init__(
            self,
            csum: _c_dm.uint32 | None = ...,
            flags: _c_dm.uint32 | None = ...,
            blocknr: _c_dm.uint64 | None = ...,
            uuid: __cs__.Array[_c_dm.uint8] | None = ...,
            magic: _c_dm.uint64 | None = ...,
            version: _c_dm.uint32 | None = ...,
            time: _c_dm.uint32 | None = ...,
            trans_id: _c_dm.uint64 | None = ...,
            held_root: _c_dm.uint64 | None = ...,
            data_space_map_root: __cs__.Array[_c_dm.uint8] | None = ...,
            metadata_space_map_root: __cs__.Array[_c_dm.uint8] | None = ...,
            data_mapping_root: _c_dm.uint64 | None = ...,
            device_details_root: _c_dm.uint64 | None = ...,
            data_block_size: _c_dm.uint32 | None = ...,
            metadata_block_size: _c_dm.uint32 | None = ...,
            metadata_nr_blocks: _c_dm.uint64 | None = ...,
            compat_flags: _c_dm.uint32 | None = ...,
            compat_ro_flags: _c_dm.uint32 | None = ...,
            incompat_flags: _c_dm.uint32 | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    class disk_device_details(__cs__.Structure):
        mapped_blocks: _c_dm.uint64
        transaction_id: _c_dm.uint64
        creation_time: _c_dm.uint32
        snapshotted_time: _c_dm.uint32
        @overload
        def __init__(
            self,
            mapped_blocks: _c_dm.uint64 | None = ...,
            transaction_id: _c_dm.uint64 | None = ...,
            creation_time: _c_dm.uint32 | None = ...,
            snapshotted_time: _c_dm.uint32 | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

# Technically `c_dm` is an instance of `_c_dm`, but then we can't use it in type hints
c_dm: TypeAlias = _c_dm

SECTOR_SIZE: int



from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime  # noqa: TC003
from functools import cache
from types import UnionType
from typing import TYPE_CHECKING, BinaryIO, get_args, get_origin, get_type_hints

from dissect.util import ts
from dissect.util.stream import MappingStream

from dissect.volume.dm.thin import ThinPool
from dissect.volume.exceptions import LVM2Error
from dissect.volume.lvm.c_lvm2 import SECTOR_SIZE
from dissect.volume.lvm.physical import LVM2Device  # noqa: TC001

if TYPE_CHECKING:
    from dissect.util.stream import RunlistStream


@dataclass(init=False)
class MetaBase:
    @classmethod
    def from_dict(cls, obj: dict, name: str | None = None, parent: MetaBase | None = None) -> MetaBase:
        inst = cls()
        inst._from_dict(obj, name=name, parent=parent)
        return inst

    def _from_dict(self, obj: dict, name: str | None = None, parent: MetaBase | None = None) -> None:
        for field_name, field_type in get_type_hints(self.__class__).items():
            if field_name.startswith("_"):
                continue

            type_ = get_origin(field_type)

            if type_ is UnionType and field_type == field_type | None:
                value = obj.get(field_name)
                field_type = get_args(field_type)[0]
                type_ = get_origin(field_type)
            else:
                value = obj[field_name]

            if type_ is dict and issubclass((value_type := get_args(field_type)[1]), MetaBase):
                value = {k: value_type.from_dict(v, name=k, parent=self) for k, v in (value or {}).items()}

            setattr(self, field_name, value)


@dataclass(init=False)
class VolumeGroup(MetaBase):
    id: str
    seqno: int
    status: list[str]
    flags: list[str]
    extent_size: int
    max_lv: int
    max_pv: int
    physical_volumes: dict[str, PhysicalVolume]
    logical_volumes: dict[str, LogicalVolume]

    system_id: str | None
    allocation_policy: str | None
    profile: str | None
    metadata_copies: int | None
    tags: list[str] | None
    historical_logical_volumes: dict[str, HistoricalLogicalVolume] | None
    format: str | None
    lock_type: str | None
    lock_args: str | None

    # Internal fields
    _name: str

    def __repr__(self) -> str:
        return f"<VolumeGroup name={self.name} id={self.id}>"

    @property
    def name(self) -> str:
        return self._name

    @property
    def pv(self) -> list[PhysicalVolume]:
        return list(self.physical_volumes.values())

    @property
    def lv(self) -> list[LogicalVolume]:
        return list(self.logical_volumes.values())

    def attach(self, devices: dict[str, LVM2Device]) -> None:
        for pv in self.physical_volumes.values():
            pv._dev = devices.get(pv.id.replace("-", ""))

    def _from_dict(self, obj: dict, name: str | None = None, parent: MetaBase | None = None) -> None:
        super()._from_dict(obj, name=name, parent=parent)
        self._name = name


@dataclass(init=False)
class PhysicalVolume(MetaBase):
    id: str
    status: list[str]
    pe_start: int
    pe_count: int

    dev_size: int | None
    device: str | None
    device_id: str | None
    device_id_type: str | None
    ba_start: int | None
    ba_size: int | None
    tags: list[str] | None

    # Internal fields
    _name: str
    _volume_group: VolumeGroup
    _dev: LVM2Device | None

    def __repr__(self) -> str:
        return f"<PhysicalVolume name={self.name} id={self.id}>"

    @property
    def name(self) -> str:
        return self._name

    @property
    def vg(self) -> VolumeGroup:
        return self._volume_group

    @property
    def volume_group(self) -> VolumeGroup:
        return self._volume_group

    @property
    def dev(self) -> LVM2Device | None:
        return self._dev

    @property
    def size(self) -> int:
        """Return the size of the physical volume in bytes.

        This can sometimes be larger than what the LVM2Device dictates, in which case the VolumeGroup overwrites it:
            https://github.com/lvmteam/lvm2/blob/6e208b81ec587434097de3207fbd1ecd7a0afb8c/lib/format_text/layout.h#L44

        From the code, this is also the value the pvck tool uses to update pv_header.device_size_xl.
        """
        # This size is also displayed when showing physical device information with pvdisplay:
        #   https://github.com/lvmteam/lvm2/blob/6e208b81ec587434097de3207fbd1ecd7a0afb8c/lib/display/display.c#L291
        if self.dev_size:
            return self.dev_size * self.vg.extent_size
        # The size of the stripes on this device + size of the metadata
        return (self.pe_start + self.pe_count * self.vg.extent_size) * SECTOR_SIZE

    def _from_dict(self, obj: dict, name: str | None = None, parent: MetaBase | None = None) -> None:
        super()._from_dict(obj, name=name, parent=parent)
        self._name = name
        self._volume_group = parent

    def open(self) -> BinaryIO:
        """Opens the physical device.

        Automatically uses the corect size as determined by :attr:`size`
        """
        if self.dev is None:
            raise LVM2Error(f"Physical volume not found: {self.name} (id={self.id}, device={self.device})")

        stream: RunlistStream = self.dev.open()
        if stream.size >= self.size:
            return stream

        size = max(stream.size, self.size)
        area = self.dev._data_area_descriptors[0]

        stream.runlist = [(area.offset // SECTOR_SIZE, size // SECTOR_SIZE)]
        stream.size = size
        return stream


@dataclass(init=False)
class LogicalVolume(MetaBase):
    id: str
    status: list[str]
    flags: list[str]
    segment_count: int

    creation_time: datetime | None
    creation_host: str | None
    lock_args: str | None
    allocation_policy: str | None
    profile: str | None
    read_ahead: int | None
    tags: list[str] | None

    # Internal fields
    _name: str
    _volume_group: VolumeGroup
    _segments: list[Segment]

    def __repr__(self) -> str:
        return f"<LogicalVolume name={self.name} id={self.id} segments={len(self.segments)}>"

    @property
    def name(self) -> str:
        return self._name

    @property
    def vg(self) -> VolumeGroup:
        return self._volume_group

    @property
    def volume_group(self) -> VolumeGroup:
        return self._volume_group

    @property
    def segments(self) -> list[Segment]:
        return self._segments

    @property
    def is_visible(self) -> bool:
        return "VISIBLE" in self.status

    @property
    def type(self) -> str | None:
        return self.segments[0].type if len(self.segments) else None

    def open(self) -> BinaryIO:
        stream = MappingStream()
        extent_size = self.volume_group.extent_size * SECTOR_SIZE
        for segment in self.segments:
            offset = segment.start_extent * extent_size
            size = segment.extent_count * extent_size
            stream.add(offset, size, segment.open())
        return stream

    def _from_dict(self, obj: dict, name: str | None = None, parent: MetaBase | None = None) -> None:
        super()._from_dict(obj, name=name, parent=parent)
        self._name = name
        self._volume_group = parent
        self._segments = [Segment.from_dict(v, name=k, parent=self) for k, v in obj.items() if isinstance(v, dict)]


@dataclass(init=False)
class HistoricalLogicalVolume(MetaBase):
    id: str
    name: str | None
    creation_time: datetime | None
    removal_time: datetime | None
    origin: str | None
    descendants: list[str] | None

    # Internal fields
    _volume_group: VolumeGroup

    def __repr__(self) -> str:
        return f"<HistoricalLogicalVolume name={self.name} id={self.id} removal_time={self.removal_time}>"

    @property
    def vg(self) -> VolumeGroup:
        return self._volume_group

    @property
    def volume_group(self) -> VolumeGroup:
        return self._volume_group

    def _from_dict(self, obj: dict, name: str | None = None, parent: MetaBase | None = None) -> None:
        super()._from_dict(obj, name=name, parent=parent)
        self.name = self.name or name
        self._volume_group = parent

        if self.creation_time:
            self.creation_time = ts.from_unix(self.creation_time)

        if self.removal_time:
            self.removal_time = ts.from_unix(self.removal_time)


@dataclass(init=False)
class Segment(MetaBase):
    start_extent: int
    extent_count: int
    type: str

    reshape_count: int | None
    data_copies: int | None
    tags: list[str] | None

    # Internal fields
    _name: str
    _logical_volume: LogicalVolume
    _flags: list[str]

    def __repr__(self) -> str:
        fields = (
            f"start_extent={self.start_extent} extent_count={self.extent_count} type={self.type} flags={self.flags}"
        )
        return f"<{self.__class__.__name__} {fields}>"

    @property
    def name(self) -> str:
        return self._name

    @property
    def lv(self) -> LogicalVolume:
        return self._logical_volume

    @property
    def logical_volume(self) -> LogicalVolume:
        return self._logical_volume

    @property
    def flags(self) -> list[str]:
        return self._flags

    def open(self) -> BinaryIO:
        raise NotImplementedError(f"{self.__class__.__name__} is not implemented yet")

    @classmethod
    def from_dict(cls, obj: dict, name: str | None = None, parent: MetaBase | None = None) -> Segment:
        SEGMENT_CLASS_MAP = {
            "linear": StripedSegment,
            "striped": StripedSegment,
            "mirror": MirrorSegment,
            "snapshot": SnapshotSegment,
            "thin": ThinSegment,
            "thin-pool": ThinPoolSegment,
            "cache": CacheSegment,
            "cache-pool": CachePoolSegment,
            "writecache": WriteCacheSegment,
            "integrity": IntegritySegment,
            "error": ErrorSegment,
            "free": FreeSegment,
            "zero": ZeroSegment,
            "vdo": VdoSegment,
            "vdo-pool": VdoPoolSegment,
            "raid": RAIDSegment,
            "raid0": RAIDSegment,
            "raid0_meta": RAIDSegment,
            "raid1": RAIDSegment,
            "raid10": RAIDSegment,
            "raid10_near": RAIDSegment,
            "raid4": RAIDSegment,
            "raid5": RAIDSegment,
            "raid5_n": RAIDSegment,
            "raid5_la": RAIDSegment,
            "raid5_ls": RAIDSegment,
            "raid5_ra": RAIDSegment,
            "raid5_rs": RAIDSegment,
            "raid6": RAIDSegment,
            "raid6_nc": RAIDSegment,
            "raid6_nr": RAIDSegment,
            "raid6_zr": RAIDSegment,
            "raid6_la_6": RAIDSegment,
            "raid6_ls_6": RAIDSegment,
            "raid6_ra_6": RAIDSegment,
            "raid6_rs_6": RAIDSegment,
            "raid6_n_6": RAIDSegment,
        }

        type = obj["type"].split("+", 1)[0]
        return super(Segment, SEGMENT_CLASS_MAP.get(type, Segment)).from_dict(obj, name=name, parent=parent)

    def _from_dict(self, obj: dict, name: str | None = None, parent: MetaBase | None = None) -> None:
        super()._from_dict(obj, name=name, parent=parent)
        self.type, *self._flags = self.type.split("+")
        self._name = name
        self._logical_volume = parent


@dataclass(init=False)
class StripedSegment(Segment):
    stripe_count: int
    stripe_size: int | None
    stripes: list[tuple[str, int]]

    def open(self) -> BinaryIO:
        stream = MappingStream()
        pv = self._logical_volume._volume_group.physical_volumes
        extent_size = self._logical_volume._volume_group.extent_size * SECTOR_SIZE

        opened_pv = {}

        offset = 0
        for pv_name, extent_offset in self.stripes:
            if pv_name not in opened_pv:
                pv_fh = pv[pv_name].open()
                opened_pv[pv_name] = pv_fh
            else:
                pv_fh = opened_pv[pv_name]

            stripe_size = (self.stripe_size or self.extent_count) * extent_size
            stream.add(offset, stripe_size, pv_fh, extent_offset * extent_size)
            offset += stripe_size

        return stream

    def _from_dict(self, obj: dict, name: str | None = None, parent: MetaBase | None = None) -> None:
        super()._from_dict(obj, name=name, parent=parent)
        self.stripes = [tuple(self.stripes[i : i + 2]) for i in range(0, len(self.stripes), 2)]


@dataclass(init=False)
class MirrorSegment(Segment):
    mirror_count: int

    extents_moved: int | None
    region_size: int | None
    mirror_log: str | None
    mirrors: list[tuple[str, int]] | None

    def open(self) -> BinaryIO:
        # Just open the first mirror we can
        for lv_name, _ in self.mirrors:
            try:
                lv = self._logical_volume._volume_group.logical_volumes[lv_name]
                return lv.open()
            except Exception:  # noqa: PERF203
                pass
        else:
            raise LVM2Error("No mirrors available")

    def _from_dict(self, obj: dict, name: str | None = None, parent: MetaBase | None = None) -> None:
        super()._from_dict(obj, name=name, parent=parent)

        if self.mirrors:
            self.mirrors = [tuple(self.mirrors[i : i + 2]) for i in range(0, len(self.mirrors), 2)]


@dataclass(init=False)
class SnapshotSegment(Segment):
    chunk_size: int

    merging_store: str | None
    cow_store: str | None
    origin: str | None


@dataclass(init=False)
class ThinSegment(Segment):
    thin_pool: str
    transaction_id: int

    origin: str | None
    merge: str | None
    device_id: int | None
    external_origin: str | None

    def open(self) -> BinaryIO:
        extent_size = self._logical_volume._volume_group.extent_size * SECTOR_SIZE

        thin_pool_lv = self._logical_volume._volume_group.logical_volumes[self.thin_pool]
        thin_pool_segment: ThinPoolSegment = thin_pool_lv.segments[0]
        thin_pool = thin_pool_segment.open_pool()
        return thin_pool.open(self.device_id, self.extent_count * extent_size)


@dataclass(init=False)
class ThinPoolSegment(Segment):
    metadata: str
    pool: str
    transaction_id: int
    chunk_size: int

    discards: str | None
    zero_new_blocks: int | None
    crop_metadata: int | None

    def __init__(self):
        self.open_pool = cache(self.open_pool)

    def open(self) -> BinaryIO:
        raise RuntimeError("Opening a thin-pool for reading is not possible, use open_pool() instead")

    def open_pool(self) -> ThinPool:
        lvs = self._logical_volume._volume_group.logical_volumes
        return ThinPool(lvs[self.metadata].open(), lvs[self.pool].open())


@dataclass(init=False)
class CacheSegment(Segment):
    cache_pool: str
    origin: str

    cleaner: int | None
    chunk_size: int | None
    cache_mode: str | None
    policy: str | None
    policy_settings: dict[str, int] | None
    metadata_format: int | None
    metadata_start: int | None
    metadata_len: int | None
    data_start: int | None
    data_len: int | None
    metadata_id: str | None
    data_id: str | None


@dataclass(init=False)
class CachePoolSegment(Segment):
    data: str
    metadata: str

    metadata_format: int | None
    chunk_size: int | None
    cache_mode: str | None
    policy: str | None
    policy_settings: dict[str, int] | None


@dataclass(init=False)
class WriteCacheSegment(Segment):
    origin: str
    writecache: str
    writecache_block_size: int

    high_watermark: int | None
    low_watermark: int | None
    writeback_jobs: int | None
    autocommit_blocks: int | None
    autocommit_time: int | None
    fua: int | None
    nofua: int | None
    cleaner: int | None
    max_age: int | None
    metadata_only: int | None
    pause_writeback: int | None
    writecache_setting_key: str | None
    writecache_setting_val: str | None


@dataclass(init=False)
class IntegritySegment(Segment):
    origin: str
    data_sectors: int
    mode: str
    tag_size: int
    block_size: int
    internal_hash: str

    meta_dev: str | None
    recalculate: int | None
    journal_sectors: int | None
    interleave_sectors: int | None
    buffer_sectors: int | None
    journal_watermark: int | None
    commit_time: int | None
    bitmap_flush_interval: int | None
    sectors_per_bit: int | None


@dataclass(init=False)
class ErrorSegment(Segment):
    pass


@dataclass(init=False)
class FreeSegment(Segment):
    pass


@dataclass(init=False)
class ZeroSegment(Segment):
    pass


@dataclass(init=False)
class VdoSegment(Segment):
    vdo_pool: str
    vdo_offset: int


@dataclass(init=False)
class VdoPoolSegment(Segment):
    data: str
    header_size: int
    virtual_extents: int
    use_compression: bool
    use_deduplication: bool
    use_metadata_hints: bool
    minimum_io_size: int
    block_map_cache_size_mb: int
    block_map_era_length: int
    use_sparse_index: int
    index_memory_size_mb: int
    max_discard: int
    slab_size_mb: int
    ack_threads: int
    bio_threads: int
    bio_rotation: int
    cpu_threads: int
    hash_zone_threads: int
    logical_threads: int
    physical_threads: int

    write_policy: str | None


@dataclass(init=False)
class RAIDSegment(Segment):
    device_count: int
    stripe_count: int
    region_size: int
    stripe_size: int
    writebehind: int
    min_recovery_rate: int
    max_recovery_rate: int

    data_copies: int | None
    data_offset: int | None
    raids: list[str] | None
    raid0_lvs: list[str] | None

    def _from_dict(self, obj: dict, name: str | None = None, parent: MetaBase | None = None) -> None:
        super()._from_dict(obj, name=name, parent=parent)

        if self.raids:
            self.raids = [tuple(self.raids[i : i + 2]) for i in range(0, len(self.raids), 2)]



from __future__ import annotations

from enum import IntEnum
from typing import TYPE_CHECKING, NamedTuple

from dissect.util.stream import AlignedStream, MappingStream

from dissect.volume.exceptions import RAIDError

if TYPE_CHECKING:
    from dissect.volume.raid.raid import DiskMap, PhysicalDisk, VirtualDisk


class Level(IntEnum):
    """RAID level identifiers."""

    LINEAR = -1
    # The rest is really just the RAID number
    RAID0 = 0
    RAID1 = 1
    RAID3 = 3
    RAID4 = 4
    RAID5 = 5
    RAID6 = 6
    RAID10 = 10


class Layout(IntEnum):
    """RAID layout identifiers.

    "Coincidentally" these align with Linux MD.
    """

    LEFT_ASYMMETRIC = 0  # Rotating Parity N with Data Restart
    RIGHT_ASYMMETRIC = 1  # Rotating Parity 0 with Data Restart
    LEFT_SYMMETRIC = 2  # Rotating Parity N with Data Continuation
    RIGHT_SYMMETRIC = 3  # Rotating Parity 0 with Data Continuation
    PARITY_0 = 4  # P or P,Q are initial devices
    PARITY_N = 5  # P or P,Q are final devices
    ROTATING_ZERO_RESTART = 8  # DDF PRL=6 RLQ=1
    ROTATING_N_RESTART = 9  # DDF PRL=6 RLQ=2
    ROTATING_N_CONTINUE = 10  # DDF PRL=6 RLQ=3
    LEFT_ASYMMETRIC_6 = 16
    RIGHT_ASYMMETRIC_6 = 17
    LEFT_SYMMETRIC_6 = 18
    RIGHT_SYMMETRIC_6 = 19
    PARITY_0_6 = 20
    PARITY_N_6 = PARITY_N


class LinearStream(MappingStream):
    """Implements a stream on a linear RAID set."""

    def __init__(self, virtual_disk: VirtualDisk):
        super().__init__()
        self.virtual_disk = virtual_disk
        self.disk_map: DiskMap = dict(sorted(virtual_disk.disk_map.items()))

        if len(self.disk_map) != virtual_disk.num_disks:
            raise RAIDError(f"Missing disks in linear RAID set {virtual_disk.uuid} ({virtual_disk.name})")

        offset = 0
        for disk_offset, disk in self.disk_map.values():
            self.add(offset, disk.size, disk.open(), disk_offset)
            offset += disk.size


class Zone(NamedTuple):
    zone_end: int
    dev_start: int
    devices: DiskMap


class RAID0Stream(AlignedStream):
    """Implements a stream on a RAID0 set."""

    def __init__(self, virtual_disk: VirtualDisk):
        self.virtual_disk = virtual_disk
        self.disk_map = dict(sorted(self.virtual_disk.disk_map.items()))

        if len(self.disk_map) != virtual_disk.num_disks:
            raise RAIDError(f"Missing disks in RAID0 set {virtual_disk.uuid} ({virtual_disk.name})")

        # Determine how many strip zones we need to construct
        # If a RAID0 set consists of devices with different sizes, additional strip zones
        # may exist on the larger devices but not on the smaller ones
        # Reference: create_strip_zones
        rounded_sizes: dict[PhysicalDisk, int] = {}
        stripe_size = virtual_disk.stripe_size
        num_strip_zones = 0
        for idx1, (_, dev1) in self.disk_map.items():
            rounded_sizes[dev1] = (dev1.size // stripe_size) * stripe_size

            has_same_size = False
            # Check if dev1 is unequal in size to the sizes of any of the previous devices
            # If so, this means an extra strip zone is present
            for idx2, (_, dev2) in self.disk_map.items():
                if idx1 == idx2:
                    break

                if rounded_sizes[dev1] == dev2.size:
                    has_same_size = True
                    break

            if not has_same_size:
                num_strip_zones += 1

        # Determine the smallest device
        smallest = None
        for _, dev in self.disk_map.values():
            if not smallest or rounded_sizes[dev] < rounded_sizes[smallest]:
                smallest = dev

        # Construct the strip zones
        zones = [Zone(rounded_sizes[smallest] * len(self.disk_map), 0, self.disk_map)]

        cur_zone_end = zones[0].zone_end
        for _ in range(1, num_strip_zones):
            zone_devices = {}
            dev_start = rounded_sizes[smallest]
            smallest = None

            # Look for the next smallest device, that is: the smallest device that is larger than the "dev_start" device
            for disk_idx, (data_offset, dev) in self.disk_map.items():
                if rounded_sizes[dev] <= dev_start:
                    continue

                zone_devices[disk_idx] = (data_offset, dev)
                if not smallest or rounded_sizes[dev] < rounded_sizes[smallest]:
                    smallest = dev

            num_dev = len(zone_devices)
            cur_size = (rounded_sizes[smallest] - dev_start) * num_dev

            cur_zone_end += cur_size

            zones.append(Zone(cur_zone_end, dev_start, zone_devices))

        self.zones = zones

        super().__init__(self.virtual_disk.size)

    def _find_zone(self, offset: int) -> tuple[Zone, int] | None:
        """Return the zone and the offset within that zone a given ``offset`` is in."""
        for i, zone in enumerate(self.zones):
            if offset < zone.zone_end:
                if i:
                    offset = offset - self.zones[i - 1].zone_end
                return zone, offset
        return None, None

    def _read(self, offset: int, length: int) -> bytes:
        result = []

        stripe_size = self.virtual_disk.stripe_size
        while length:
            zone, offset_in_zone = self._find_zone(offset)
            if zone is None:
                break

            read_offset = offset_in_zone
            if len(self.zones) == 1 or len(self.zones[1].devices) == 1:
                read_offset = offset

            stripe, offset_in_stripe = divmod(read_offset, stripe_size)
            chunk = offset_in_zone // (stripe_size * len(zone.devices))

            offset_in_device = (chunk * stripe_size) + offset_in_stripe
            device_start, device = zone.devices[stripe % len(zone.devices)]

            stripe_remaining = stripe_size - offset_in_stripe
            read_length = min(length, stripe_remaining)

            offset_on_disk = device.offset + device_start + offset_in_device
            device.fh.seek(offset_on_disk)
            result.append(device.fh.read(read_length))

            length -= read_length
            offset += read_length

        return b"".join(result)


class RAID456Stream(AlignedStream):
    """Implements a stream on a RAID5 set."""

    def __init__(self, virtual_disk: VirtualDisk):
        self.virtual_disk = virtual_disk
        self.level = self.virtual_disk.level
        self.algorithm = self.virtual_disk.layout
        self.max_degraded = 2 if self.level == 6 else 1

        self.disk_map = self.virtual_disk.disk_map
        if len(self.disk_map) < self.virtual_disk.num_disks - self.max_degraded:
            raise RAIDError(f"Missing disks in RAID{self.level} set {virtual_disk.uuid} ({virtual_disk.name})")

        super().__init__(self.virtual_disk.size, self.virtual_disk.stripe_size)

    def _get_stripe_read_info(self, offset: int) -> tuple[int, int, int, int, int | None]:
        """Calculate the stripe, offset in the stripe, data disk, parity disk and "Q" parity disk for a given sector."""
        # Reference: raid5_compute_sector
        stripe_size = self.virtual_disk.stripe_size
        raid_disks = self.virtual_disk.num_disks
        data_disks = raid_disks - self.max_degraded

        stripe_number, offset_in_stripe = divmod(offset, stripe_size)
        stripe, dd_idx = divmod(stripe_number, data_disks)

        pd_idx = None
        qd_idx = None
        ddf_layout = False

        if self.level == 4:
            pd_idx = data_disks

        elif self.level == 5:
            if self.algorithm == Layout.LEFT_ASYMMETRIC:
                pd_idx = data_disks - (stripe % raid_disks)
                if dd_idx >= pd_idx:
                    dd_idx += 1

            elif self.algorithm == Layout.RIGHT_ASYMMETRIC:
                pd_idx = stripe % raid_disks
                if dd_idx >= pd_idx:
                    dd_idx += 1

            elif self.algorithm == Layout.LEFT_SYMMETRIC:
                pd_idx = data_disks - (stripe % raid_disks)
                dd_idx = (pd_idx + 1 + dd_idx) % raid_disks

            elif self.algorithm == Layout.RIGHT_SYMMETRIC:
                pd_idx = stripe % raid_disks
                dd_idx = (pd_idx + 1 + dd_idx) % raid_disks

            elif self.algorithm == Layout.PARITY_0:
                pd_idx = 0
                dd_idx += 1

            elif self.algorithm == Layout.PARITY_N:
                pd_idx = data_disks

            else:
                raise RAIDError(f"Invalid RAID algorithm: {self.algorithm}")

        elif self.level == 6:
            if self.algorithm == Layout.LEFT_ASYMMETRIC:
                pd_idx = raid_disks - 1 - (stripe % raid_disks)
                qd_idx = pd_idx + 1
                if pd_idx == raid_disks - 1:
                    # Q D D D P
                    dd_idx += 1
                    qd_idx = 0
                elif dd_idx >= pd_idx:
                    # D D P Q D
                    dd_idx += 2

            elif self.algorithm == Layout.RIGHT_ASYMMETRIC:
                pd_idx = stripe % raid_disks
                qd_idx = pd_idx + 1
                if pd_idx == raid_disks - 1:
                    # Q D D D P
                    dd_idx += 1
                    qd_idx = 0
                elif dd_idx >= pd_idx:
                    # D D P Q D
                    dd_idx += 2

            elif self.algorithm == Layout.LEFT_SYMMETRIC:
                pd_idx = raid_disks - 1 - (stripe % raid_disks)
                qd_idx = (pd_idx + 1) % raid_disks
                dd_idx = (pd_idx + 2 + dd_idx) % raid_disks

            elif self.algorithm == Layout.RIGHT_SYMMETRIC:
                pd_idx = stripe % raid_disks
                qd_idx = (pd_idx + 1) % raid_disks
                dd_idx = (pd_idx + 2 + dd_idx) % raid_disks

            elif self.algorithm == Layout.PARITY_0:
                pd_idx = 0
                qd_idx = 1
                dd_idx += 2

            elif self.algorithm == Layout.PARITY_N:
                pd_idx = data_disks
                qd_idx = data_disks + 1

            elif self.algorithm == Layout.ROTATING_ZERO_RESTART:
                # Exactly the same as RIGHT_ASYMMETRIC, but or of blocks for computing Q is different
                pd_idx = stripe % raid_disks
                qd_idx = pd_idx + 1
                if pd_idx == raid_disks - 1:
                    dd_idx += 1
                    qd_idx = 0
                elif dd_idx >= pd_idx:
                    # D D P Q D
                    dd_idx += 2
                ddf_layout = True

            elif self.algorithm == Layout.ROTATING_N_RESTART:
                # Same as left_asymmetric, but first stripe is
                # D D D P Q  rather than
                # Q D D D P
                pd_idx = raid_disks - 1 - ((stripe + 1) % raid_disks)
                qd_idx = pd_idx + 1
                if pd_idx == raid_disks - 1:
                    # Q D D D P
                    dd_idx += 1
                    qd_idx = 0
                elif dd_idx >= pd_idx:
                    # D D P Q D
                    dd_idx += 2
                ddf_layout = True

            elif self.algorithm == Layout.ROTATING_N_CONTINUE:
                # Same as left_symmetric but Q is before P
                pd_idx = raid_disks - 1 - (stripe % raid_disks)
                qd_idx = (pd_idx + raid_disks - 1) % raid_disks
                dd_idx = (pd_idx + 1 + dd_idx) % raid_disks
                ddf_layout = True

            elif self.algorithm == Layout.LEFT_ASYMMETRIC_6:
                # RAID5 left_asymmetric, with Q on last device
                pd_idx = data_disks - (stripe % (raid_disks - 1))
                if dd_idx >= pd_idx:
                    dd_idx += 1
                qd_idx = raid_disks - 1

            elif self.algorithm == Layout.RIGHT_ASYMMETRIC_6:
                pd_idx = stripe % (raid_disks - 1)
                if dd_idx >= pd_idx:
                    dd_idx += 1
                    qd_idx = raid_disks - 1

            elif self.algorithm == Layout.LEFT_SYMMETRIC_6:
                pd_idx = data_disks - (stripe % (raid_disks - 1))
                dd_idx = (pd_idx + 1 + dd_idx) % (raid_disks - 1)
                qd_idx = raid_disks - 1

            elif self.algorithm == Layout.RIGHT_SYMMETRIC_6:
                pd_idx = stripe % (raid_disks - 1)
                dd_idx = (pd_idx + 1 + dd_idx) % (raid_disks - 1)
                qd_idx = raid_disks - 1

            elif self.algorithm == Layout.PARITY_0_6:
                pd_idx = 0
                dd_idx += 1
                qd_idx = raid_disks - 1
            else:
                raise RAIDError(f"Invalid RAID algorithm: {self.algorithm}")
        else:
            raise RAIDError(f"Invalid RAID level: {self.level}")

        return stripe, offset_in_stripe, dd_idx, pd_idx, qd_idx, ddf_layout

    def _read(self, offset: int, length: int) -> bytes:
        result = []

        stripe_size = self.virtual_disk.stripe_size
        while length:
            stripe, offset_in_stripe, dd_idx, _, _, _ = self._get_stripe_read_info(offset)
            offset_in_device = stripe * stripe_size + offset_in_stripe
            dd_start, dd_dev = self.disk_map[dd_idx]

            stripe_remaining = stripe_size - offset_in_stripe
            read_length = min(length, stripe_remaining)

            offset_on_disk = dd_dev.offset + dd_start + offset_in_device
            dd_dev.fh.seek(offset_on_disk)
            result.append(dd_dev.fh.read(length))

            length -= read_length
            offset += read_length

        return b"".join(result)


class RAID10Stream(AlignedStream):
    """Implements a stream on a RAID10 set."""

    def __init__(self, virtual_disk: VirtualDisk):
        self.virtual_disk = virtual_disk
        self.raid_disks = self.virtual_disk.num_disks
        self.disk_map = virtual_disk.disk_map

        # Reference: setup_geo
        layout = virtual_disk.layout
        self.near_copies = layout & 0xFF
        self.far_copies = (layout >> 8) & 0xFF
        self.far_offset = layout & (1 << 16)

        use_far_sets = layout >> 17
        if use_far_sets == 0:
            # Original layout
            self.far_set_size = self.raid_disks
        elif use_far_sets == 1:
            # "Improved" but bugged layout
            self.far_set_size = self.raid_disks // self.far_copies
        elif use_far_sets == 2:
            # "Improved" and fixed layout
            self.far_set_size = self.far_copies * self.near_copies
        else:
            raise ValueError("Invalid RAID10 layout: {layout:#x}")

        self.last_far_set_start = ((self.raid_disks / self.far_set_size) - 1) * self.far_set_size
        self.last_far_set_size = self.far_set_size + (self.raid_disks % self.far_set_size)

        super().__init__(self.virtual_disk.size, self.virtual_disk.stripe_size)

    def _read(self, offset: int, length: int) -> bytes:
        result = []

        stripe_size = self.virtual_disk.stripe_size
        while length:
            # Reference: __raid10_find_phys
            stripe_number, offset_in_stripe = divmod(offset, stripe_size)
            stripe, dev = divmod(stripe_number * self.near_copies, self.raid_disks)

            if self.far_offset:
                stripe *= self.far_copies
            device_start, device = self.disk_map[dev]

            stripe_remaining = stripe_size - offset_in_stripe
            read_length = min(length, stripe_remaining)

            offset_on_disk = device.offset + device_start + (stripe * stripe_size) + offset_in_stripe
            device.fh.seek(offset_on_disk)
            result.append(device.fh.read(length))

            length -= read_length
            offset += read_length

        return b"".join(result)



from __future__ import annotations

from functools import lru_cache
from typing import BinaryIO

from dissect.volume.dm.c_dm import SECTOR_SIZE, c_dm


class BTree:
    def __init__(self, fh: BinaryIO, root: int, block_size: int):
        self.fh = fh
        self.root = root
        self.block_size = block_size
        self._block_size_bytes = self.block_size * SECTOR_SIZE

        self._read_node = lru_cache(64)(self._read_node)

    def _read_node(self, block: int) -> Node:
        self.fh.seek(block * self._block_size_bytes)
        return Node(self.fh.read(self._block_size_bytes))

    def lookup(self, keys: int | list[int], want_high: bool = False) -> bytes | None:
        keys = [keys] if not isinstance(keys, list) else keys

        root = self.root
        last_level = len(keys) - 1
        value = None
        for i, key in enumerate(keys):
            found_key, value = self._lookup(root, key, want_high)
            if found_key is None or found_key != key:
                return None

            if i < last_level:
                root = int.from_bytes(value, "little")

        return value

    def _lookup(self, root: int, key: int, want_high: bool = False) -> tuple[int | None, bytes | None]:
        block = root
        while True:
            node = self._read_node(block)

            low = -1
            high = node.num_entries
            while high - low > 1:
                mid = low + ((high - low) // 2)
                cmp_key = node.key(mid)

                if cmp_key == key:
                    result = mid
                    break

                if cmp_key < key:
                    low = mid
                else:
                    high = mid
            else:
                result = high if want_high else low

            if result < 0 or result >= node.num_entries:
                return None, None

            if node.is_internal:
                block = int.from_bytes(node.value(result), "little")
            elif node.is_leaf:
                return node.key(result), node.value(result)


class Node:
    def __init__(self, buf: bytes):
        self.buf = memoryview(buf)
        self.header = c_dm.node_header(self.buf)

        self.num_entries = self.header.nr_entries
        self.max_entries = self.header.max_entries
        self.value_size = self.header.value_size

        key_area_start = len(c_dm.node_header)
        key_area_size = self.max_entries * 8
        key_area_end = key_area_start + key_area_size
        value_area_size = self.max_entries * self.value_size

        self._key_area = self.buf[key_area_start:key_area_end]
        self._value_area = self.buf[key_area_end : key_area_end + value_area_size]

        self.key = lru_cache(1024)(self.key)

    @property
    def is_internal(self) -> bool:
        return bool(self.header.flags & c_dm.node_flags.INTERNAL_NODE.value)

    @property
    def is_leaf(self) -> bool:
        return bool(self.header.flags & c_dm.node_flags.LEAF_NODE.value)

    def key(self, idx: int) -> int:
        if idx >= self.num_entries:
            raise IndexError("Key index out of bounds")
        area = idx * 8
        return int.from_bytes(self._key_area[area : area + 8], "little")

    def value(self, idx: int) -> bytes:
        if idx >= self.num_entries:
            raise IndexError("Value index out of bounds")
        area = idx * self.value_size
        return self._value_area[area : area + self.value_size]



from __future__ import annotations

from dissect.volume.disk.disk import Disk
from dissect.volume.disk.partition import Partition
from dissect.volume.exceptions import DiskError

__all__ = [
    "Disk",
    "DiskError",
    "Partition",
]



from __future__ import annotations

from typing import TYPE_CHECKING, BinaryIO
from uuid import UUID

from dissect.cstruct import cstruct

from dissect.volume.disk.partition import Partition
from dissect.volume.disk.schemes.mbr import MBR
from dissect.volume.exceptions import DiskError

if TYPE_CHECKING:
    from collections.abc import Iterator

gpt_def = """
// http://en.wikipedia.org/wiki/GUID_Partition_Table
struct GPT_HEADER {
    char        signature[8];
    uint32      revision;
    uint32      header_size;
    uint32      crc32;
    uint32      reserved;
    uint64      current_lba;
    uint64      backup_lba;
    uint64      first_usable_lba;
    uint64      last_usable_lba;
    char        guid[16];
    uint64      lba_partition_array;
    uint32      partition_table_count;
    uint32      partition_entry_size;
    uint32      partition_table_crc;
    char        _[420];
};

struct GPT_PARTITION {
    char        type_guid[16];
    char        partition_guid[16];
    uint64      first_lba;
    uint64      last_lba;
    uint64      attribute_flags;
    char        name[72];   // UTF16 encoded
};

// 0 (0x00)     16 bytes     Partition type GUID
// 16 (0x10)    16 bytes     Unique partition GUID
// 32 (0x20)    8 bytes      First LBA (little endian)
// 40 (0x28)    8 bytes      Last LBA (inclusive, usually odd)
// 48 (0x30)    8 bytes      Attribute flags (e.g. bit 60 denotes read-only)
// 56 (0x38)    72 bytes     Partition name (36 UTF-16LE code units)
"""

c_gpt = cstruct().load(gpt_def)


class GPT:
    """GUID Partition Table."""

    def __init__(self, fh: BinaryIO, sector_size: int = 512):
        self.fh = fh
        self.sector_size = sector_size

        self.mbr = MBR(fh, sector_size=sector_size)
        gpt_parts = [p for p in self.mbr.partitions if p.type == 0xEE]
        if not gpt_parts:
            raise DiskError("Invalid protective/hybrid MBR, could not find 0xEE GPT partition")

        gpt_part = gpt_parts[0]
        fh.seek(gpt_part.offset)

        self.gpt = c_gpt.GPT_HEADER(fh)
        if self.gpt.signature != b"EFI PART":
            raise DiskError(f"Invalid GPT signature, expected 'EFI PART', got {self.gpt.signature!r}.")

        self.partitions: list[Partition] = list(self._partitions())

    def _partitions(self) -> Iterator[Partition]:
        # First iterate MBR partitions
        # When we find the GPT partition, iterate GPT partitions
        # This should support Hybrid GPT
        seen_gpt = False

        for mbr_part in self.mbr.partitions:
            # There can be multiple protective MBR entries
            if mbr_part.type == 0xEE:
                if seen_gpt:
                    continue

                base_offset = self.gpt.lba_partition_array * self.sector_size
                self.fh.seek(base_offset)

                # Numbering a hybrid GPT is a bit weird, I think it's best to do nothing
                # This way the MBR and the GPT have their own separate partition numbers
                for part_num in range(self.gpt.partition_table_count):
                    self.fh.seek(base_offset + (part_num * self.gpt.partition_entry_size))

                    partition = c_gpt.GPT_PARTITION(self.fh)
                    if partition.first_lba == 0:
                        continue

                    name = (
                        partition.name.decode("utf-16-le", "ignore")
                        .split("\x00")[0]
                        .strip("\x00")
                        .strip("\uffff")  # a non-character in UTF-16
                    )

                    yield Partition(
                        disk=self,
                        number=part_num + 1,  # partitions are 1-indexed
                        offset=partition.first_lba * self.sector_size,
                        size=(partition.last_lba - partition.first_lba) * self.sector_size,
                        type=UUID(bytes_le=partition.type_guid),
                        name=name,
                        flags=partition.attribute_flags,
                        guid=UUID(bytes_le=partition.partition_guid),
                        raw=partition,
                    )

                seen_gpt = True
            else:
                yield mbr_part
